"""Django template language based expression engine."""
from django import template
from django.conf import settings

from resolwe.flow.exprengines import BaseExpressionEngine
from resolwe.flow.models import Data, hydrate_input_references, hydrate_input_uploads, render_template


class ExpressionEngine(BaseExpressionEngine):

    """Represents a workflow expression engine."""

    def eval(self, data):
        """Evaluate the script and return executable."""
        try:
            inputs = data.input.copy()
            hydrate_input_references(inputs, data.process.input_schema)
            hydrate_input_uploads(inputs, data.process.input_schema)

            inputs['proc'] = {
                'data_id': data.id,
                'data_path': settings.FLOW_EXECUTOR['DATA_PATH'],
            }

            script_template = data.process.run.get('bash', '')

            script = render_template(script_template, template.Context(inputs))

        except template.TemplateSyntaxError as ex:
            data.status = Data.STATUS_ERROR
            data.process_error.append('Error in process script: {}'.format(ex))
            data.save()
            return None

        return script



"""
==========
Data Purge
==========

"""
import os
import shutil

from django.conf import settings

from flow.models import Data, iterate_fields, Storage


def data_purge(data_ids=None, delete=False, verbosity=1):
    """Print files not referenced from meta data.

    If data_ids not given, run on all data objects.
    If delete is True, delete unreferenced files.

    """
    data_path = settings.FLOW_EXECUTOR['data_path']
    path_offset = len(data_path) + 24 + 2

    def remove_file(fn, paths):
        """From paths remove fn and dirs before fn in dir tree."""
        while fn:
            for i in range(len(paths) - 1, -1, -1):
                if fn == paths[i]:
                    paths.pop(i)
            fn, _ = os.path.split(fn)

    def remove_tree(fn, paths):
        """From paths remove fn and dirs before or after fn in dir tree."""
        for i in range(len(paths) - 1, -1, -1):
            head = paths[i]
            while head:
                if fn == head:
                    paths.pop(i)
                    break
                head, _ = os.path.split(head)

        remove_file(fn, paths)

    def subfiles(root):
        """Extend unreferenced list with all subdirs and files in top dir."""
        subs = []
        for path, dirs, files in os.walk(root, topdown=False):
            path = path[path_offset:]
            subs.extend(os.path.join(path, f) for f in files)
            subs.extend(os.path.join(path, d) for d in dirs)
        return subs

    check = Data.objects.all()
    purge = []
    ref_storage = set()
    finished = [Data.STATUS_DONE, Data.STATUS_ERROR]
    if data_ids:
        check = check.filter(id__in=data_ids)

        purge = set(data_ids)
    else:
        purge = set(os.listdir(data_path))

    # Remove not referenced
    purge = purge.difference(str(d.id) for d in check.only('pk'))
    purge = [os.path.join(data_path, p) for p in purge]

    check = check.filter(status__in=finished)

    for d in check:
        root = os.path.join(data_path, str(d.id))
        subs = subfiles(root)

        meta_fields = [
            ['output', 'output_schema'],
            ['static', 'static_schema'],
            ['var', 'var_template'],
        ]

        for meta_field, meta_field_schema in meta_fields:
            for field_schema, fields in iterate_fields(d[meta_field], d[meta_field_schema]):
                if 'type' in field_schema:
                    field_type = field_schema['type']
                    field_name = field_schema['name']

                    # Remove basic:file: entries
                    if field_type.startswith('basic:file:'):
                        remove_file(fields[field_name]['file'], subs)

                    # Remove list:basic:file: entries
                    elif field_type.startswith('list:basic:file:'):
                        for field in fields[field_name]:
                            remove_file(field['file'], subs)

                    # Remove basic:url: entries
                    elif field_type.startswith('basic:url:'):
                        remove_file(fields[field_name]['url'], subs)

                    # Remove list:basic:url: entries
                    elif field_type.startswith('list:basic:url:'):
                        for field in fields[field_name]:
                            remove_file(field['url'], subs)

                    # Remove refs entries
                    if field_type.startswith('basic:file:') or field_type.startswith('basic:url:'):
                        for ref in fields[field_name].get('refs', []):
                            remove_tree(ref, subs)

                    elif field_type.startswith('list:basic:file:') or field_type.startswith('list:basic:url:'):
                        for field in fields[field_name]:
                            for ref in field.get(['refs'], []):
                                remove_tree(ref, subs)

                    if field_schema['type'].startswith('basic:json:'):
                        ref_storage.add(fields[field_name])

        purge.extend(os.path.join(root, f) for f in subs)

    # IMPORTANT: Unreferenced `Storage` objects can be determined only
    #            if all `Data` objects are checked.
    unref_storage = Storage.objects.none()
    if (verbosity or delete) and data_ids is None:
        unref_storage = Storage.objects.filter(pk__nin=ref_storage).only('pk')

    if verbosity >= 1:
        # Print unreferenced files
        if purge:
            print "Unreferenced files ({}):".format(len(purge))
            for name in purge:
                print "  {}".format(name)
        else:
            print "No unreferenced files"

        if unref_storage:
            print "Unreferenced Storage objects ({}):".format(unref_storage.count())
            for s in unref_storage:
                print "  {}".format(s.pk)
        elif data_ids is None:
            print "No unreferenced Storage objects"

    # Go through unreferenced files and delete them
    if delete:
        for name in purge:
            if os.path.isfile(name) or os.path.islink(name):
                os.remove(name)
            elif os.path.isdir(name):
                shutil.rmtree(name)

        # Safety check to prevent deleting all `Storage` objects
        if data_ids is None:
            unref_storage.delete()



"""
.. autoclass:: resolwe.flow.tests.BaseProcessorTestCase
   :members:

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import hashlib
import gzip
import json
import os
import shutil
import stat
import zipfile
try:
    from itertools import filterfalse  # py3
except ImportError:
    from itertools import ifilterfalse as filterfalse  # py2

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import management
from django.test import TestCase, override_settings
from django.utils.crypto import get_random_string
from django.utils.text import slugify

from resolwe.flow.models import Data, dict_dot, iterate_fields, Collection, DescriptorSchema, Process, Storage
from resolwe.flow.managers import manager


SCHEMAS_FIXTURE_CACHE = None


def _register_schemas():
    """Register process and descriptor schemas.

    Process and DescriptorSchema definitions are registered when first
    test is callled and cached to SCHEMAS_FIXTURE_CACHE global variable

    """
    Process.objects.all().delete()

    global SCHEMAS_FIXTURE_CACHE  # pylint: disable=global-statement
    if SCHEMAS_FIXTURE_CACHE:
        Process.objects.bulk_create(SCHEMAS_FIXTURE_CACHE['processes'])
        DescriptorSchema.objects.bulk_create(SCHEMAS_FIXTURE_CACHE['descriptor_schemas'])
    else:
        user_model = get_user_model()

        if not user_model.objects.filter(is_superuser=True).exists():
            user_model.objects.create_superuser(username="admin", email='admin@example.com', password="admin_pass")

        management.call_command('register', force=True, testing=True, verbosity='0')

        SCHEMAS_FIXTURE_CACHE = {}
        SCHEMAS_FIXTURE_CACHE['processes'] = list(Process.objects.all())  # list forces db query execution
        SCHEMAS_FIXTURE_CACHE['descriptor_schemas'] = list(DescriptorSchema.objects.all())


flow_executor_settings = settings.FLOW_EXECUTOR.copy()
# since we don't know what uid/gid will be used inside Docker executor, others
# must have all permissions on the data directory
flow_executor_settings['DATA_DIR_MODE'] = 0o777
# create a temporary upload directory so we can give it relaxed permissions and
# make it convenient for use by Docker
flow_executor_settings['UPLOAD_PATH'] = os.path.join(flow_executor_settings['UPLOAD_PATH'], 'test_upload')

# replace existing Docker UPLOAD_PATH mapping with the new upload directory
flow_docker_mappings = copy.copy(getattr(settings, "FLOW_DOCKER_MAPPINGS", {}))
for map_ in flow_docker_mappings:
    if settings.FLOW_EXECUTOR['UPLOAD_PATH'] in map_['src']:
        map_['src'] = flow_executor_settings['UPLOAD_PATH']


@override_settings(FLOW_EXECUTOR=flow_executor_settings)
@override_settings(FLOW_DOCKER_MAPPINGS=flow_docker_mappings)
class ProcessTestCase(TestCase):

    """Base class for writing processor tests.

    This class is subclass of Django's ``TestCase`` with some specific
    functions used for testing processors.

    To write a processor test use standard Django's syntax for writing
    tests and follow next steps:

    #. Put input files (if any) in ``tests/files`` folder of a Django
       application.
    #. Run test with run_processor.
    #. Check if processor has finished successfully with
       assertDone function.
    #. Assert processor's output with :func:`assertFiles`,
       :func:`assertFields` and :func:`assertJSON` functions.

    .. note::
        When creating a test case for a custom Django application,
        subclass this class and over-ride the ``self.files_path`` with:

        .. code-block::
            self.files_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')

    .. DANGER::
        If output files don't exist in ``tests/files`` folder of a
        Django application, they are created automatically.
        But you have to check that they are correct before using them
        for further runs.

    """

    def setUp(self):
        super(ProcessTestCase, self).setUp()
        self.admin = get_user_model().objects.create_superuser(
            username="admin", email='admin@example.com', password="admin_pass")
        _register_schemas()

        self.collection = Collection.objects.create(contributor=self.admin)
        self.files_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')
        self._keep_all = False
        self._keep_failed = False

        # create temporary upload dir
        self.upload_path = settings.FLOW_EXECUTOR['UPLOAD_PATH']
        self.upload_dir_existed = os.path.isdir(self.upload_path)
        if self.upload_dir_existed:
            # since we don't know what uid/gid will be used inside Docker executor,
            # others must have all permissions on the upload directory
            upload_dir_mode = stat.S_IMODE(os.stat(self.upload_path).st_mode)
            others_all_perm = stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH
            if upload_dir_mode & others_all_perm != others_all_perm:
                raise ValueError("Incorrect permissions ({}) for upload dir ({}). "
                                 "Change it so that others will have read, write and execute "
                                 "permissions.".format(oct(upload_dir_mode), self.upload_path))
        else:
            os.mkdir(self.upload_path)
            # os.mkdir is not guaranteed to set the given mode
            os.chmod(self.upload_path, 0o777)

    def tearDown(self):
        super(ProcessTestCase, self).tearDown()

        # Delete Data objects and their files unless keep_failed
        for d in Data.objects.all():
            if self._keep_all or (self._keep_failed and d.status == "error"):
                print("KEEPING DATA: {}".format(d.pk))
            else:
                data_dir = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(d.pk))
                d.delete()
                shutil.rmtree(data_dir, ignore_errors=True)

        # remove temporary upload dir
        if not self.upload_dir_existed and not self._keep_all and not self._keep_failed:
            shutil.rmtree(self.upload_path, ignore_errors=True)

    def keep_all(self):
        """Do not delete output files after test for all data."""
        self._keep_all = True

    def keep_failed(self):
        """Do not delete output files after test for failed data."""
        self._keep_failed = True

    def run_processor(self, *args, **kwargs):
        return self.run_process(*args, **kwargs)
        # TODO: warning

    def run_process(self, process_slug, input_={}, assert_status=Data.STATUS_DONE, run_manager=True, verbosity=0):
        """Runs given processor with specified inputs.

        If input is file, file path should be given relative to
        ``tests/files`` folder of a Django application.
        If ``assert_status`` is given check if Data object's status
        matches ``assert_status`` after finishing processor.

        :param processor_name: name of the processor to run
        :type processor_name: :obj:`str`

        :param ``input_``: Input paramaters for processor. You don't
            have to specifie parameters for which default values are
            given.
        :type ``input_``: :obj:`dict`

        :param ``assert_status``: Desired status of Data object
        :type ``assert_status``: :obj:`str`

        :return: :obj:`resolwe.flow.models.Data` object which is created by
            the processor.

        """

        # backward compatibility
        process_slug = slugify(process_slug.replace(':', '-'))

        p = Process.objects.get(slug=process_slug)

        for field_schema, fields in iterate_fields(input_, p.input_schema):
            # copy referenced files to upload dir
            if field_schema['type'] == "basic:file:":
                for app_config in apps.get_app_configs():

                    old_path = os.path.join(app_config.path, 'tests', 'files', fields[field_schema['name']])
                    if os.path.isfile(old_path):
                        file_name = os.path.basename(fields[field_schema['name']])
                        new_path = os.path.join(self.upload_path, file_name)
                        shutil.copy2(old_path, new_path)
                        # since we don't know what uid/gid will be used inside Docker executor,
                        # we must give others read and write permissions
                        os.chmod(new_path, 0o666)
                        fields[field_schema['name']] = {
                            'file': file_name,
                            'file_temp': file_name,
                        }
                        break

            # convert primary keys to strings
            if field_schema['type'].startswith('data:'):
                fields[field_schema['name']] = str(fields[field_schema['name']])
            if field_schema['type'].startswith('list:data:'):
                fields[field_schema['name']] = [str(obj) for obj in fields[field_schema['name']]]

        d = Data.objects.create(
            input=input_,
            contributor=self.admin,
            process=p,
            slug=get_random_string(length=6))
        self.collection.data.add(d)

        if run_manager:
            manager.communicate(run_sync=True, verbosity=verbosity)

        # Fetch latest Data object from database
        d = Data.objects.get(pk=d.pk)

        if not run_manager and assert_status == Data.STATUS_DONE:
            assert_status = Data.STATUS_RESOLVING

        if assert_status:
            self.assertStatus(d, assert_status)

        return d

    def assertStatus(self, obj, status):  # pylint: disable=invalid-name
        """Check if Data object's status is 'status'.

        :param obj: Data object for which to check status
        :type obj: :obj:`resolwe.flow.models.Data`
        :param status: Data status to check
        :type status: str

        """
        self.assertEqual(obj.status, status,
                         msg="Data status is '{}', not '{}'".format(obj.status, status) +
                         self._debug_info(obj))

    def assertFields(self, obj, path, value):  # pylint: disable=invalid-name
        """Compare Data object's field to given value.

        :param obj: Data object with field to compare
        :type obj: :obj:`resolwe.flow.models.Data`

        :param path: Path to field in Data object.
        :type path: :obj:`str`

        :param value: Desired value.
        :type value: :obj:`str`

        """
        field = dict_dot(obj.output, path)
        self.assertEqual(field, value,
                         msg="Field 'output.{}' mismatch: {} != {}".format(path, field, str(value)) +
                         self._debug_info(obj))

    def assertFiles(self, obj, field_path, fn, compression=None, filter=lambda _: False):  # pylint: disable=invalid-name
        """Compare output file of a processor to the given correct file.

        :param obj: Data object which includes file that we want to
            compare.
        :type obj: :obj:`resolwe.flow.models.Data`

        :param field_path: Path to file name in Data object.
        :type field_path: :obj:`str`

        :param fn: File name (and relative path) of file to which we
            want to compare. Name/path is relative to ``tests/files``
            folder of a Django application.
        :type fn: :obj:`str`

        :param compression: If not None, files will be uncompressed with
            the appropriate compression library before comparison.
            Currently supported compression formats are "gzip" and
            "zip".
        :type compression: :obj:`str`

        :param filter: Function for filtering the contents of output files. It
            is used in :obj:`itertools.filterfalse` function and takes one
            parameter, a line of the output file. If it returns `True`, the
            line is excluded from comparison of the two files.
        :type filter: :obj:`function`

        """
        open_kwargs = {}
        if compression is None:
            open_fn = open
            # by default, open() will open files as text and return str
            # objects, but we need bytes objects
            open_kwargs['mode'] = 'rb'
        elif compression == 'gzip':
            open_fn = gzip.open
        elif compression == 'zip':
            open_fn = zipfile.ZipFile.open
        else:
            raise ValueError("Unsupported compression format.")

        field = dict_dot(obj.output, field_path)
        output = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(obj.pk), field['file'])
        with open_fn(output, **open_kwargs) as output_file:
            output_contents = b"".join([line for line in filterfalse(filter, output_file)])
        output_hash = hashlib.sha256(output_contents).hexdigest()

        wanted = os.path.join(self.files_path, fn)

        if not os.path.isfile(wanted):
            shutil.copyfile(output, wanted)
            self.fail(msg="Output file {} missing so it was created.".format(fn))

        with open_fn(wanted, **open_kwargs) as wanted_file:
            wanted_contents = b"".join([line for line in filterfalse(filter, wanted_file)])
        wanted_hash = hashlib.sha256(wanted_contents).hexdigest()
        self.assertEqual(wanted_hash, output_hash,
                         msg="File contents hash mismatch: {} != {}".format(
                             wanted_hash, output_hash) + self._debug_info(obj))

    def assertFileExists(self, obj, field_path):  # pylint: disable=invalid-name
        """Compare output file of a processor to the given correct file.

        :param obj: Data object which includes file that we want to
            compare.
        :type obj: :obj:`resolwe.flow.models.Data`

        :param field_path: Path to file name in Data object.
        :type field_path: :obj:`str`

        """
        field = dict_dot(obj.output, field_path)
        output = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(obj.pk), field['file'])

        if not os.path.isfile(output):
            self.fail(msg="File {} does not exist.".format(field_path))

    def assertJSON(self, obj, storage, field_path, file_name):  # pylint: disable=invalid-name
        """Compare JSON in Storage object to the given correct output.

        :param obj: Data object which includes file that we want to
            compare.
        :type obj: :obj:`resolwe.flow.models.Data`

        :param storage: Storage (or storage id) which contains JSON to
            compare.
        :type storage: :obj:`resolwe.flow.models.Storage` or :obj:`str`

        :param field_path: Path to JSON subset to compare in Storage
            object. If it is empty, entire Storage object will be
            compared.
        :type field_path: :obj:`str`

        :param file_name: File name (and relative path) of file to
            which we want to compare. Name/path is relative to
            ``tests/files`` folder of a Django application.
        :type file_name: :obj:`str`

        """
        self.assertEqual(os.path.splitext(file_name)[1], '.gz', msg='File extension must be .gz')

        if not isinstance(storage, Storage):
            storage = Storage.objects.get(pk=storage)

        storage_obj = dict_dot(storage.json, field_path)

        file_path = os.path.join(self.files_path, file_name)
        if not os.path.isfile(file_path):
            with gzip.open(file_path, 'w') as f:
                json.dump(storage_obj, f)

            self.fail(msg="Output file {} missing so it was created.".format(file_name))

        with gzip.open(file_path) as f:
            file_obj = json.load(f)

        self.assertEqual(storage_obj, file_obj,
                         msg="Storage {} field '{}' does not match file {}".format(
                             storage.id, field_path, file_name) + self._debug_info(obj))

    def _debug_info(self, data):
        """Return data's debugging information."""
        msg_header = "Debugging information for data object {}".format(data.pk)
        msg = "\n\n" + len(msg_header)*"=" + "\n" + msg_header + "\n" + len(msg_header)*"=" + "\n"
        path = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(data.pk), "stdout.txt")
        if os.path.isfile(path):
            msg += "\nstdout.txt:\n" + 11*"-" + "\n"
            with open(path, 'r') as fn:
                msg += fn.read()

        if data.process_error:
            msg += "\nProcess' errors:\n" + 16*"-" + "\n"
            msg += "\n".join(data.process_error)

        if data.process_warning:
            msg += "\nProcess' warnings:\n" + 18*"-" + "\n"
            msg += "\n".join(data.process_warning)

        return msg



"""
================
Flow Serializers
================

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from rest_framework import serializers, status
from rest_framework.exceptions import APIException
from rest_framework.fields import empty
from resolwe.flow.models import Process, Collection, Data, DescriptorSchema, Trigger, Storage


class NoContentError(APIException):
    status_code = status.HTTP_204_NO_CONTENT
    detail = 'The content has not changed'


class ResolweBaseSerializer(serializers.ModelSerializer):
    """Base serializer for all `Resolwe` objects.

    This class is inherited from `django_rest_framework`'s
    `ModelSerialzer` class. The difference is that
    `update_protected_fields` are removed from `data` dict when update
    is performed.

    To check whether the class is called to create an instance or
    to update an existing one, it checks its value. If the value is
    `None`, a new instance is being created.
    The `update_protected_fields` tuple can be defined in the `Meta`
    class of child class.

    `NoContentError` is raised if no data would be changed, so we
    prevent changing `modified` field.

    """
    def __init__(self, instance=None, data=empty, **kwargs):
        if (instance is not None and data is not empty and
                hasattr(self.Meta, 'update_protected_fields')):
            for field in self.Meta.update_protected_fields:
                if field in data:
                    data.pop(field)

            # prevent changing `modified` field if no field would be changed
            if set(data.keys()).issubset(set(self.Meta.read_only_fields)):
                raise NoContentError()

        return super(ResolweBaseSerializer, self).__init__(instance, data, **kwargs)


class ProcessSerializer(ResolweBaseSerializer):

    """Serializer for Process objects."""

    class Meta:
        """ProcessSerializer Meta options."""
        model = Process
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'slug', 'name', 'created', 'modified', 'version', 'type',
                            'category', 'persistence', 'description', 'input_schema',
                            'output_schema', 'run', )
        fields = update_protected_fields + read_only_fields


class DescriptorSchemaSerializer(ResolweBaseSerializer):

    """Serializer for DescriptorSchema objects."""

    class Meta:
        """TemplateSerializer Meta options."""
        model = DescriptorSchema
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'version', 'schema') + update_protected_fields + read_only_fields


class DataSerializer(ResolweBaseSerializer):

    """Serializer for Data objects."""

    process_name = serializers.CharField(source='process.name', read_only=True)
    process_type = serializers.CharField(source='process.type', read_only=True)
    process_input_schema = serializers.JSONField(source='process.input_schema', read_only=True)
    process_output_schema = serializers.JSONField(source='process.output_schema', read_only=True)
    descriptor_schema = DescriptorSchemaSerializer(required=False, read_only=True)

    name = serializers.CharField(read_only=False, required=False)
    slug = serializers.CharField(read_only=False, required=False)

    class Meta:
        """DataSerializer Meta options."""
        model = Data
        update_protected_fields = ('contributor', 'process',)
        read_only_fields = ('id', 'created', 'modified', 'started', 'finished', 'checksum',
                            'status', 'process_progress', 'process_rc', 'process_info',
                            'process_warning', 'process_error', 'process_type',
                            'process_input_schema', 'process_output_schema',
                            'process_name')
        fields = ('slug', 'name', 'contributor', 'input', 'output', 'descriptor_schema',
                  'descriptor') + update_protected_fields + read_only_fields


class CollectionSerializer(ResolweBaseSerializer):

    """Serializer for Collection objects."""

    descriptor_schema = DescriptorSchemaSerializer(required=False)

    slug = serializers.CharField(read_only=False, required=False)

    class Meta:
        """CollectionSerializer Meta options."""
        model = Collection
        update_protected_fields = ('contributor',)
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'description', 'settings', 'descriptor_schema', 'descriptor',
                  'data') + update_protected_fields + read_only_fields

    def __init__(self, *args, **kwargs):
        super(CollectionSerializer, self).__init__(*args, **kwargs)

        request = kwargs.get('context', {}).get('request', None)

        if request and request.query_params.get('hydrate_data', False):
            self.fields['data'] = DataSerializer(many=True, read_only=True)
        else:
            self.fields['data'] = serializers.PrimaryKeyRelatedField(many=True, read_only=True)


class TriggerSerializer(ResolweBaseSerializer):

    """Serializer for Trigger objects."""

    class Meta:
        """TriggerSerializer Meta options."""
        model = Trigger
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'trigger', 'trigger_input', 'process', 'input', 'collection',
                  'autorun') + update_protected_fields + read_only_fields


class StorageSerializer(ResolweBaseSerializer):

    """Serializer for Storage objects."""

    class Meta:
        """StorageSerializer Meta options."""
        model = Storage
        update_protected_fields = ('contributor', )
        read_only_fields = ('id', 'created', 'modified')
        fields = ('slug', 'name', 'data', 'json') + update_protected_fields + read_only_fields



# -*- coding: utf-8 -*-

import sys
import os
import shlex

base_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "..")

# Get package metadata from 'resolwe/__about__.py' file
about = {}
with open(os.path.join(base_dir, 'resolwe', '__about__.py')) as f:
    exec(f.read(), about)

# -- General configuration ------------------------------------------------

# The extension modules to enable.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Resolwe'
version = about['__version__']
release = version
author = about['__author__']
copyright = about['__copyright__']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'CHANGELOG.rst']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = 'sphinx_rtd_theme'

# Output file base name for HTML help builder.
htmlhelp_basename = 'Resolwedoc'

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}



"""Workflow Expression Engines"""


class BaseExpressionEngine(object):

    """Represents a workflow expression engine."""

    def eval(self, data_id, script):
        """Evaluate the script and return executable."""
        raise NotImplementedError('subclasses of BaseExpressionEngine may require an eval() method')



from __future__ import absolute_import, division, print_function, unicode_literals
"""Workflow Executors"""

import json
import logging
import os
import six

from django.apps import apps
from django.conf import settings

if settings.USE_TZ:
    from django.utils.timezone import now
else:
    import datetime  # pylint: disable=wrong-import-order
    now = datetime.datetime.now  # pylint: disable=invalid-name

from resolwe.flow.models import Data, dict_dot
from resolwe.utils import BraceMessage as __


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
CWD = os.getcwd()


def iterjson(text):
    """Decode JSON stream."""
    decoder = json.JSONDecoder()
    while len(text) > 0:
        obj, ndx = decoder.raw_decode(text)

        if not isinstance(obj, dict):
            raise ValueError()

        text = text[ndx:].lstrip('\r\n')
        yield obj


class BaseFlowExecutor(object):

    """Represents a workflow executor

    .. WARNING::
        Ensure this class can be serialized since some engines will
        use it this way (e.g. Celery).
    """

    def get_tools(self):
        tools_paths = []
        for app_config in apps.get_app_configs():
            proc_path = os.path.join(app_config.path, 'tools')
            if os.path.isdir(proc_path):
                tools_paths.append(proc_path)
        return tools_paths

    def start(self):
        pass

    def run_script(self, script):
        raise NotImplementedError('`run_script` function must be defined')

    def end(self):
        pass

    def get_stdout(self):
        return self.stdout

    def update_data_status(self, **kwargs):
        data = Data.objects.get(pk=self.data_id)
        for key, value in kwargs.items():
            setattr(data, key, value)

        # Ensure that we only update the fields that were changed.
        data.save(update_fields=kwargs.keys())

    def run(self, data_id, script, verbosity=1):
        """Execute the script and save results."""
        if verbosity >= 1:
            print('RUN: {} {}'.format(data_id, script))

        self.data_id = data_id

        dir_mode = settings.FLOW_EXECUTOR.get('DATA_DIR_MODE', 0o755)

        output_path = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(data_id))
        os.mkdir(output_path)
        # os.mkdir is not guaranteed to set the given mode
        os.chmod(output_path, dir_mode)
        os.chdir(output_path)

        log_file = open('stdout.txt', 'w+')
        json_file = open('jsonout.txt', 'w+')

        proc_pid = self.start()

        self.update_data_status(
            status=Data.STATUS_PROCESSING,
            started=now(),
            process_pid=proc_pid
        )

        # Run processor and handle intermediate results
        self.run_script(script)
        spawn_processors = []
        output = {}
        process_error, process_warning, process_info = [], [], []
        process_progress, process_rc = 0, 0

        # read processor output
        try:
            stdout = self.get_stdout()
            while True:
                line = stdout.readline()
                if not line:
                    break

                try:
                    if line.strip().startswith('run'):
                        # Save processor and spawn if no errors
                        log_file.write(line)
                        log_file.flush()

                        for obj in iterjson(line[3:].strip()):
                            spawn_processors.append(obj)
                    else:
                        # If JSON, save to MongoDB
                        updates = {}
                        for obj in iterjson(line):
                            for key, val in six.iteritems(obj):
                                if key.startswith('proc.'):
                                    if key == 'proc.error':
                                        process_error.append(val)
                                        if not process_rc:
                                            process_rc = 1
                                            updates['process_rc'] = process_rc
                                        updates['process_error'] = process_error
                                        updates['status'] = Data.STATUS_ERROR
                                    elif key == 'proc.warning':
                                        process_warning.append(val)
                                        updates['process_warning'] = process_warning
                                    elif key == 'proc.info':
                                        process_info.append(val)
                                        updates['process_info'] = process_info
                                    elif key == 'proc.rc':
                                        process_rc = int(val)
                                        updates['process_rc'] = process_rc
                                        if process_rc != 0:
                                            updates['status'] = Data.STATUS_ERROR
                                    elif key == 'proc.progress':
                                        process_progress = int(float(val) * 100)
                                        updates['process_progress'] = process_progress
                                else:
                                    dict_dot(output, key, val)
                                    updates['output'] = output

                        if updates:
                            updates['modified'] = now()
                            self.update_data_status(**updates)

                        if process_rc > 0:
                            log_file.close()
                            json_file.close()
                            os.chdir(CWD)
                            return

                        # Debug output
                        # Not referenced in Data object
                        json_file.write(line)
                        json_file.flush()

                except ValueError as ex:
                    # Ignore if not JSON
                    log_file.write(line)
                    log_file.flush()

        except MemoryError as ex:
            logger.error(__("Out of memory: {}", ex))

        except IOError as ex:
            # TODO: if ex.errno == 28: no more free space
            raise ex
        finally:
            # Store results
            log_file.close()
            json_file.close()
            os.chdir(CWD)

        return_code = self.end()

        if process_rc < return_code:
            process_rc = return_code

        if process_rc == 0:
            self.update_data_status(
                status=Data.STATUS_DONE,
                process_progress=100,
                finished=now()
            )
        else:
            self.update_data_status(
                status=Data.STATUS_ERROR,
                process_progress=100,
                process_rc=process_rc,
                finished=now()
            )

        # try:
        #     # Cleanup after processor
        #     data_purge(data_ids=[data_id], delete=True, verbosity=0)
        # except:  # pylint: disable=bare-except
        #     logger.error(__("Purge error:\n\n{}", traceback.format_exc()))

        # if not update_data(data):  # Data was deleted
        #     # Restore original directory
        #     os.chdir(settings.PROJECT_ROOT)
        #     return

        # if data.status == Data.STATUS_DONE and spawn_processors:
        #     # Spawn processors
        #     os.chdir(HOME)
        #     for d in spawn_processors:
        #         d['case_ids'] = [case_id]
        #         d['author_id'] = data.author_id
        #         data = Data(**d)
        #         data.save()

        #         run_triggers(data_id=data.id, autorun=True)

        # Restore original directory
        # os.chdir(settings.PROJECT_ROOT)

        # return data_id

    def terminate(self, data_id):
        """Terminate a running script."""
        raise NotImplementedError('subclasses of BaseFlowExecutor may require a terminate() method')



from __future__ import absolute_import, division, print_function, unicode_literals

import os

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string


class BaseProcessesFinder(object):
    """
    A base processes loader to be used for custom staticfiles finder
    classes.
    """

    def find_processes(self):
        raise NotImplementedError('subclasses of BaseProcessesLoader must provide a find_processes() method')

    def find_descriptors(self):
        raise NotImplementedError('subclasses of BaseProcessesLoader must provide a find_descriptors() method')


class FileSystemProcessesFinder(BaseProcessesFinder):
    def find_processes(self):
        return getattr(settings, 'FLOW_PROCESSES_DIRS', ())

    def find_descriptors(self):
        return getattr(settings, 'FLOW_DESCRIPTORS_DIRS', ())


class AppDirectoriesFinder(BaseProcessesFinder):
    def _find_folders(self, folder_name):
        found_folders = []
        for app_config in apps.get_app_configs():
            folder_path = os.path.join(app_config.path, folder_name)
            if os.path.isdir(folder_path):
                found_folders.append(folder_path)
        return found_folders

    def find_processes(self):
        return self._find_folders('processes')

    def find_descriptors(self):
        return self._find_folders('descriptors')


def get_finders():
    for finder_path in settings.FLOW_PROCESSES_FINDERS:
        yield get_finder(finder_path)


def get_finder(import_path):
    Finder = import_string(import_path)
    if not issubclass(Finder, BaseProcessesFinder):
        raise ImproperlyConfigured(
            'Finder "{}" is not a subclass of "{}"'.format(Finder, BaseProcessesFinder))
    return Finder()



from __future__ import absolute_import, division, print_function, unicode_literals

from .base import BaseManager


class Manager(BaseManager):

    def run(self, data_id, script, run_sync=False, verbosity=1):
        self.executor.run(data_id, script, verbosity=verbosity)



"""
==================
Flow Configuration
==================

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class FlowConfig(AppConfig):

    """Flow AppConfig."""

    name = 'resolwe.flow'
    verbose_name = _("Resolwe Dataflow")



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-01-06 05:36
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import versionfield


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Data',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100)),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('started', models.DateTimeField(blank=True, null=True)),
                ('finished', models.DateTimeField(blank=True, null=True)),
                ('checksum', models.CharField(blank=True, max_length=40, null=True, validators=[django.core.validators.RegexValidator(code='invalid_checksum', message='Checksum is exactly 40 alphanumerics', regex='^[0-9a-f]{40}$')])),
                ('status', models.CharField(choices=[('UP', 'Uploading'), ('RE', 'Resolving'), ('WT', 'Waiting'), ('PR', 'Processing'), ('OK', 'Done'), ('ER', 'Error'), ('DR', 'Dirty')], default='RE', max_length=2)),
                ('process_pid', models.PositiveIntegerField(blank=True, null=True)),
                ('process_progress', models.PositiveSmallIntegerField(default=0)),
                ('process_rc', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('process_info', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=[], size=None)),
                ('process_warning', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=[], size=None)),
                ('process_error', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=[], size=None)),
                ('input', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('output', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('descriptor', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'default_permissions': (),
                'abstract': False,
                'permissions': (('view_data', 'Can view data'), ('edit_data', 'Can edit data'), ('share_data', 'Can share data'), ('download_data', 'Can download files from data')),
            },
        ),
        migrations.CreateModel(
            name='DescriptorSchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100)),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('description', models.TextField(blank=True)),
                ('schema', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'default_permissions': (),
                'abstract': False,
                'permissions': (('view_descriptorschema', 'Can view descriptor schema'), ('edit_descriptorschema', 'Can edit descriptor schema'), ('share_descriptorschema', 'Can share descriptor schema')),
            },
        ),
        migrations.CreateModel(
            name='Process',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100)),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('type', models.CharField(max_length=100, validators=[django.core.validators.RegexValidator(code='invalid_type', message='Type may be alphanumerics separated by colon', regex='^data:[a-z0-9:]+:$')])),
                ('category', models.CharField(default='other', max_length=200, validators=[django.core.validators.RegexValidator(code='invalid_category', message='Category may be alphanumerics separated by colon', regex='^([a-z0-9]+[:\\-])*[a-z0-9]+:$')])),
                ('persistence', models.CharField(choices=[('RAW', 'Raw'), ('CAC', 'Cached'), ('TMP', 'Temp')], default='RAW', max_length=3)),
                ('priority', models.CharField(choices=[('NO', 'Normal'), ('HI', 'High')], default='NO', max_length=2)),
                ('description', models.TextField(default='')),
                ('input_schema', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=[])),
                ('output_schema', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=[])),
                ('run', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'default_permissions': (),
                'abstract': False,
                'permissions': (('view_process', 'Can view process'), ('share_process', 'Can share process')),
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100)),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('description', models.TextField(blank=True)),
                ('settings', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('data', models.ManyToManyField(to='flow.Data')),
                ('public_processes', models.ManyToManyField(to='flow.Process')),
            ],
            options={
                'default_permissions': (),
                'abstract': False,
                'permissions': (('view_project', 'Can view project'), ('edit_project', 'Can edit project'), ('share_project', 'Can share project'), ('download_project', 'Can download files from project'), ('add_project', 'Can add data objects to project')),
            },
        ),
        migrations.CreateModel(
            name='Storage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100)),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('json', django.contrib.postgres.fields.jsonb.JSONField()),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('data', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Data')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Trigger',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100)),
                ('version', versionfield.VersionField(default=0)),
                ('name', models.CharField(max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('type', models.CharField(max_length=100, validators=[django.core.validators.RegexValidator(code='invalid_type', message='Type may be alphanumerics separated by colon', regex='^data:[a-z0-9:]+:$')])),
                ('trigger', models.CharField(max_length=500)),
                ('trigger_input', models.CharField(max_length=100)),
                ('input', django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ('autorun', models.BooleanField(default=False)),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('process', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='flow.Process')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='flow.Project')),
            ],
            options={
                'default_permissions': (),
                'abstract': False,
                'permissions': (('view_trigger', 'Can view trigger'), ('edit_trigger', 'Can edit trigger'), ('share_trigger', 'Can share trigger')),
            },
        ),
        migrations.AddField(
            model_name='data',
            name='descriptor_schema',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema'),
        ),
        migrations.AddField(
            model_name='data',
            name='process',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='flow.Process'),
        ),
        migrations.AlterUniqueTogether(
            name='trigger',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='storage',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='project',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='process',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='descriptorschema',
            unique_together=set([('slug', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='data',
            unique_together=set([('slug', 'version')]),
        ),
    ]



"""
=========================
Delete Unreferenced Files
=========================

"""
from django.core.management.base import BaseCommand

from server.models import data_purge


class Command(BaseCommand):

    """Purge files with no reference in Data objects."""

    help = "Purge files with no reference in Data objects."

    def add_arguments(self, parser):
        parser.add_argument('-d', '--data', type=str, nargs='+', help="list of data ids")
        parser.add_argument('-f', '--force', action='store_true', help="delete unreferenced files")

    def handle(self, *args, **options):
        """Call function data_purge from models.py to purge files."""
        data_purge(options['data'], options['force'])



"""Local workflow executor"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import random
import shlex
import subprocess

from django.conf import settings

from .local import FlowExecutor as LocalFlowExecutor


class FlowExecutor(LocalFlowExecutor):

    def __init__(self, *args, **kwargs):
        super(FlowExecutor, self).__init__(*args, **kwargs)
        self.command = settings.FLOW_EXECUTOR.get('COMMAND', 'docker')

    def start(self):
        container_image = settings.FLOW_EXECUTOR['CONTAINER_IMAGE']

        if self.data_id != 'no_data_id':
            container_name = 'resolwe_{}'.format(self.data_id)
        else:
            # set random container name for tests
            rand_int = random.randint(1000, 9999)
            container_name = 'resolwe_test_{}'.format(rand_int)

        # render Docker mappings in FLOW_DOCKER_MAPPINGS setting
        mappings_template = getattr(settings, 'FLOW_DOCKER_MAPPINGS', [])
        context = {'data_id': self.data_id}
        mappings = [{key.format(**context): value.format(**context) for key, value in template.items()}
                    for template in mappings_template]
        # create mappings for tools
        # NOTE: To prevent processes tampering with tools, all tools are mounted read-only
        self.mappings_tools = [{'src': tool, 'dest': '/usr/local/bin/resolwe/{}'.format(i), 'mode': 'ro'}
                               for i, tool in enumerate(self.get_tools())]
        mappings += self.mappings_tools
        # create Docker --volume parameters from mappings
        volumes = " ".join(["--volume={src}:{dest}:{mode}".format(**map_) for map_ in mappings])

        # a login Bash shell is needed to source ~/.bash_profile
        self.proc = subprocess.Popen(
            shlex.split(
                '{} run --rm --interactive --name={} {} {} /bin/bash --login'.format(
                    self.command, container_name, volumes, container_image)),
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)

        self.stdout = self.proc.stdout

    def run_script(self, script):
        mappings = getattr(settings, 'FLOW_DOCKER_MAPPINGS', {})
        for map_ in mappings:
            script = script.replace(map_['src'], map_['dest'])
        # create a Bash command to add all the tools to PATH
        tools_paths = ':'.join([map_["dest"] for map_ in self.mappings_tools])
        add_tools_path = 'export PATH=$PATH:{}'.format(tools_paths)
        self.proc.stdin.write(os.linesep.join(['set -x', 'set +B', add_tools_path, script]) + os.linesep)
        self.proc.stdin.close()

    def end(self):
        self.proc.wait()

        return self.proc.returncode

    def terminate(self, data_id):
        subprocess.call(shlex.split('{} rm -f {}').format(self.command, str(data_id)))



# Command to run on local machine to migrate data from gendev:
# ./tests/manage.py genesis_migrate -u=genesis -p=resres --port=10117 --db-name=gendev_data --output=gendev.txt

from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime
import json

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from django.utils.text import slugify

from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm, get_groups_with_perms, get_users_with_perms

from resolwe.flow.models import Data, DescriptorSchema, Process, Collection, Storage, Trigger, dict_dot, iterate_fields


class Command(BaseCommand):

    """Migrate data from `Genesis`"""

    help = 'Migrate data from `Genesis`'

    def add_arguments(self, parser):
        parser.add_argument('-u', '--username', type=str, default='postgres', help="MongoDB username")
        parser.add_argument('-p', '--password', required=True, type=str, help="MongoDB password")
        parser.add_argument('--db-name', required=True, type=str, help="MongoDB database name")
        parser.add_argument('--host', type=str, default='localhost', help="MongoDB host")
        parser.add_argument('--port', type=int, default=27017, help="MongoDB port")
        parser.add_argument('--output', type=str, default=None, help="Output file for id mappings")
        parser.add_argument('--purge', action='store_true', help='Purge local database before migration')

    persistence_dict = {
        'raw': 'RAW',
        'cached': 'CAC',
        'temp': 'TMP',
    }

    status_dict = {
        'uploading': 'UP',
        'resolving': 'RE',
        'waiting': 'WT',
        'processing': 'PR',
        'done': 'OK',
        'error': 'ER',
        'dirty': 'DR',
    }

    storage_index = {}
    descriptor_schema_index = {}
    id_mapping = {_type: {} for _type in ['collection', 'process', 'data', 'storage', 'trigger']}
    collection_tags = {}

    missing_users = set()
    missing_collections = set()
    missing_default_collections = []
    missing_data = set()
    long_names = []
    unreferenced_storages = []
    orphan_triggers = []

    def process_slug(self, name):
        return slugify(name.replace(':', '-'))

    def get_contributor(self, contributor_id):
        user_model = get_user_model()
        try:
            return user_model.objects.get(pk=contributor_id)
        except user_model.DoesNotExist:
            return user_model.objects.filter(is_superuser=True).first()

    def migrate_permissions(self, new, old):
        content_type = ContentType.objects.get_for_model(new)

        def set_permissions(entity_type):
            model = get_user_model() if entity_type == 'users' else Group

            for entity_id in old['permissions'].get(entity_type, {}):
                try:
                    entity = model.objects.get(pk=int(entity_id))
                except ValueError:
                    # old format of user ids
                    continue
                except model.DoesNotExist:
                    self.missing_users.add(entity_id)
                    continue

                for perm in old['permissions'][entity_type][entity_id]:
                    if '{}_{}'.format(perm.lower(), content_type) == 'download_app':
                        # depricated permission
                        continue
                    assign_perm('{}_{}'.format(perm.lower(), content_type), entity, new)

        set_permissions('users')
        set_permissions('groups')

        def set_public_permissions():
            user = AnonymousUser()
            perms = old['permissions'].get('public', {})
            for perm in perms:
                if '{}_{}'.format(perm.lower(), content_type) == 'download_app':
                    # depricated permission
                    continue
                assign_perm('{}_{}'.format(perm.lower(), content_type), user, new)

        set_public_permissions()

    def migrate_collection(self, collection):
        new = Collection()
        new.name = collection[u'name']
        new.slug = collection[u'url_slug'] if u'url_slug' in collection else slugify(collection[u'name'])
        new.description = collection.get(u'description', '')
        new.contributor = self.get_contributor(collection['author_id'])
        new.settings = collection.get(u'settings', {})
        # Remove data table settings as they are incompatible with the new frontend.
        for applicationConfig in new.settings.values():
            applicationConfig['tables'] = {}
        # XXX: Django will change this on create
        new.created = collection[u'date_created']
        # XXX: Django will change this on save
        new.modified = collection[u'date_modified']
        new.save()

        self.migrate_permissions(new, collection)

        for tag in collection.get(u'tags', []):
            if tag not in self.collection_tags:
                self.collection_tags[tag] = []
            self.collection_tags[tag].append(new)

        self.id_mapping['collection'][str(collection[u'_id'])] = new.pk

    def migrate_process(self, process):
        new = Process()
        new.name = process[u'label']
        new.slug = self.process_slug(process[u'name'])
        new.version = process[u'version']
        new.type = process[u'type']
        new.description = process.get(u'description', '')
        new.contributor = self.get_contributor(process['author_id'])
        new.category = process.get(u'category', '')
        # XXX: Django will change this on create
        new.created = process[u'date_created']
        # XXX: Django will change this on save
        new.modified = process[u'date_modified']
        new.output_schema = process[u'output_schema']
        new.input_schema = process.get(u'input_schema', {})
        new.persistence = self.persistence_dict[process[u'persistence']]
        new.run['script'] = process[u'run'][u'bash']
        new.save()

        self.migrate_permissions(new, process)

        self.id_mapping['process'][str(process[u'_id'])] = new.pk

    def migrate_data(self, data):
        contributor = self.get_contributor(data[u'author_id'])

        # DESCRIPTOR SCHEMA ############################################
        ds_fields = []
        ds_fields.extend(data.get(u'static_schema', []))
        ds_fields.extend(data.get(u'var_template', []))
        ds_fields.sort(key=lambda d: d[u'name'])
        ds_fields_dumped = json.dumps(ds_fields)

        if ds_fields_dumped in self.descriptor_schema_index:
            descriptor_schema = self.descriptor_schema_index[ds_fields_dumped]
        else:
            descriptor_schema = DescriptorSchema(schema=ds_fields)
            descriptor_schema.name = 'data_{}_descriptor'.format(data[u'_id'])
            descriptor_schema.contributor = contributor
            descriptor_schema.save()

            self.descriptor_schema_index[ds_fields_dumped] = descriptor_schema

        descriptor = {}
        descriptor.update(data.get(u'static', {}))
        descriptor.update(data.get(u'var', {}))

        # PROCESS ######################################################
        if u'processor_version' not in data:
            data[u'processor_version'] = '0.0.0'

        process_slug = self.process_slug(data[u'processor_name'])
        process_version = data[u'processor_version']
        try:
            process = Process.objects.get(slug=process_slug, version=process_version)
        except Process.DoesNotExist:
            latest = Process.objects.filter(slug=process_slug).order_by('-version').first()

            if latest:
                process = Process()
                process.name = latest.name
                process.slug = latest.slug
                process.category = latest.category
                process.description = latest.description
                process.contributor = latest.contributor

                process.version = process_version
                process.type = data[u'type']
                process.output_schema = data[u'output_schema']
                process.input_schema = data.get(u'input_schema', {})
                process.persistence = self.persistence_dict[data[u'persistence']]

                process.run['script'] = 'gen-require common\ngen-error "Depricated process, use the latest version."'

                # XXX
                # process.created =
                # process.modified =

                process.save()

                # copy permissions from latest process
                for user, perms in get_users_with_perms(latest, attach_perms=True).iteritems():
                    for perm in perms:
                        assign_perm(perm, user, process)
                for group, perms in get_groups_with_perms(latest, attach_perms=True).iteritems():
                    for perm in perms:
                        assign_perm(perm, group, process)
            else:
                # Create dummy processor if there is no other version
                dummy_name = 'Dummy processor of type {}'.format(data[u'type'])
                try:
                    process = Process.objects.get(name=dummy_name)
                except Process.DoesNotExist:
                    process = Process.objects.create(
                        name=dummy_name,
                        slug='non-existent',
                        contributor=get_user_model().objects.filter(is_superuser=True).first(),
                        type=data[u'type'],
                        category='data:non-existent',
                        run={'script': {'gen-require common\ngen-error "This processor is not intendent to be run."'}},
                    )

        # DATA #########################################################
        new = Data()
        new.name = data.get(u'static', {}).get(u'name', '')
        if len(new.name) > 100:
            self.long_names.append(new.name)
            new.name = new.name[:97] + '...'
        new.status = self.status_dict[data[u'status']]
        new.process = process
        new.contributor = contributor
        new.input = data[u'input'] if u'input' in data else {}
        new.output = data[u'output']
        new.descriptor_schema = descriptor_schema
        new.descriptor = descriptor
        new.checksum = data.get(u'checksum', '')
        # XXX: Django will change this on create
        new.created = data[u'date_created']
        # XXX: Django will change this on save
        new.modified = data[u'date_modified']
        if u'date_start' in data and u'date_finish' in data:
            new.started = data[u'date_start']
            new.finished = data[u'date_finish']
        elif u'date_finish' in data:
            new.started = data[u'date_finish']
            new.finished = data[u'date_finish']
        elif u'date_start' in data:
            new.started = data[u'date_start']
            new.finished = data[u'date_start']
        else:
            new.started = datetime.fromtimestamp(0)
            new.finished = datetime.fromtimestamp(0)
        new.save()

        for case_id in data[u'case_ids']:
            try:
                collection = Collection.objects.get(pk=self.id_mapping[u'collection'][str(case_id)])
            except KeyError:
                self.missing_collections.add(str(case_id))
                continue
            collection.data.add(new)

        for field_schema, fields, path in iterate_fields(data[u'output'], data[u'output_schema'], ''):
            if 'type' in field_schema and field_schema['type'].startswith('basic:json:'):
                self.storage_index[fields[field_schema['name']]] = {
                    'id': new.pk,
                    'path': path,
                }

        self.migrate_permissions(new, data)

        self.id_mapping['data'][str(data[u'_id'])] = new.pk

        # DESCRIPTOR SCHEMA PERMISSIONS ################################
        for user in get_users_with_perms(new):
            assign_perm('view_descriptorschema', user, obj=descriptor_schema)

        for group in get_groups_with_perms(new):
            assign_perm('view_descriptorschema', group, obj=descriptor_schema)

    def migrate_data_references(self):
        def map_reference(reference):
            try:
                return self.id_mapping['data'][reference]
            except KeyError, error:
                self.missing_data.add(error.message)
                return None

        # Fix references in JSON documents in the second pass.
        for old_id, new_id in self.id_mapping['data'].items():
            data = Data.objects.get(pk=new_id)
            for field_schema, fields in iterate_fields(data.input, data.process.input_schema):
                if 'type' not in field_schema:
                    continue

                name = field_schema['name']
                value = fields[name]
                if field_schema['type'].startswith('data:'):
                    fields[name] = map_reference(value)
                elif field_schema['type'].startswith('list:data:'):
                    fields[name] = map(map_reference, value)

            data.save()

    def migrate_storage(self, storage):
        if str(storage[u'_id']) not in self.storage_index:
            self.unreferenced_storages.append(storage[u'_id'])
            return 1

        data_id = self.storage_index[str(storage[u'_id'])]['id']
        data_path = self.storage_index[str(storage[u'_id'])]['path']
        data = Data.objects.get(pk=data_id)

        new = Storage()
        new.name = 'data_{}_storage'.format(data_id)
        new.data = data
        new.json = storage[u'json']
        new.contributor = self.get_contributor(storage[u'author_id'])
        # XXX: Django will change this on create
        new.created = storage[u'date_created']
        # XXX: Django will change this on save
        new.modified = storage[u'date_modified']
        new.save()

        dict_dot(data.output, data_path, new.pk)
        data.save()

        self.id_mapping['storage'][str(storage[u'_id'])] = new.pk

    def migrate_trigger(self, trigger):
        if str(trigger[u'case_id']) not in self.id_mapping[u'collection']:
            self.orphan_triggers.append(str(trigger[u'case_id']))
            return 1

        new = Trigger()
        new.name = trigger[u'name']
        new.contributor = self.get_contributor(trigger[u'author_id'])
        new.type = trigger[u'type']
        new.trigger = trigger[u'trigger']
        new.trigger_input = trigger[u'trigger_input']
        new.process = Process.objects.filter(
            slug=self.process_slug(trigger[u'processor_name'])).order_by('-version').first()
        new.input = trigger[u'input']
        new.collection = Collection.objects.get(pk=self.id_mapping[u'collection'][str(trigger[u'case_id'])])
        new.autorun = trigger[u'autorun']
        # XXX: Django will change this on create
        new.created = trigger[u'date_created']
        # XXX: Django will change this on save
        new.modified = trigger[u'date_modified']
        new.save()

        self.id_mapping['trigger'][str(trigger[u'_id'])] = new.pk

    def clear_database(self):
        for model in [Collection, Data, Process, Storage, DescriptorSchema,
                      GroupObjectPermission, UserObjectPermission]:
            model.objects.all().delete()

    def handle(self, *args, **options):
        try:
            from pymongo import MongoClient
        except ImportError:
            self.stdout.write('PyMongo is required')
            exit(1)

        if options['purge']:
            self.clear_database()

        mongo_uri = 'mongodb://{username}:{password}@{host}:{port}/{db_name}'.format(**options)
        client = MongoClient(mongo_uri)

        self.stdout.write('Migrating collections...')
        for collection in client[options['db_name']]['case'].find():
            self.migrate_collection(collection)
        self.stdout.write('DONE')

        self.stdout.write('Migrating processes...')
        for process in client[options['db_name']].processor.find():
            self.migrate_process(process)
        self.stdout.write('DONE')

        self.stdout.write('Migrating data...')
        for data in client[options['db_name']].data.find():
            self.migrate_data(data)
        self.migrate_data_references()
        self.stdout.write('DONE')

        self.stdout.write('Migrating storage...')
        for storage in client[options['db_name']].storage.find():
            self.migrate_storage(storage)
        self.stdout.write('DONE')

        self.stdout.write('Migrating trigger...')
        for trigger in client[options['db_name']].trigger.find():
            self.migrate_trigger(trigger)
        self.stdout.write('DONE')

        if options['output']:
            with open(options['output'], 'w') as fn:
                json.dump(self.id_mapping, fn)
        else:
            self.stdout.write('\nID mappings:', json.dumps(self.id_mapping))

        self.stdout.write('\nMissing users: {}'.format(sorted(list(self.missing_users))))
        self.stdout.write('Missing collections (referenced in Data objects): {}'.format(list(self.missing_collections)))
        self.stdout.write('Missing collections (referenced in triggers): {}'.format(self.orphan_triggers))
        self.stdout.write('Missing data (referenced in Data objects): {}'.format(list(self.missing_data)))
        self.stdout.write('Number of shortened names: {}'.format(len(self.long_names)))
        self.stdout.write('Number of unreferenced Storage objects: {}'.format(len(self.unreferenced_storages)))



"""
=======
Resolwe
=======

Open source enterprise dataflow engine in Django.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from . import flow

from resolwe.__about__ import (
    __author__, __copyright__, __email__, __license__, __summary__, __title__,
    __url__, __version__,
)



"""Central place for package metadata."""

# NOTE: We use __title__ instead of simply __name__ since the latter would
#       interfere with a global variable __name__ denoting object's name.
__title__ = 'resolwe'
__summary__ = 'Open source enterprise dataflow engine in Django'
__url__ = 'https://github.com/genialis/resolwe'

# Semantic versioning is used. For more information see:
# https://packaging.python.org/en/latest/distributing/#semantic-versioning-preferred
__version__ = '1.2.0'

__author__ = 'Genialis d.o.o.'
__email__ = 'dev-team@genialis.com'

__license__ = 'Apache License (2.0)'
__copyright__ = '2015-2016, ' + __author__

__all__ = [
    "__title__", "__summary__", "__url__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]



"""
==========
Flow Views
==========

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import pkgutil
from importlib import import_module

from django.db import IntegrityError, transaction
from django.db.models import Q
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.utils._os import upath


from rest_framework import exceptions, mixins, viewsets, status
from rest_framework.decorators import detail_route, list_route
from rest_framework.response import Response

from guardian import shortcuts
import rest_framework_filters as filters

from .managers import manager
from .models import Collection, Process, Data, DescriptorSchema, Trigger, Storage
from .serializers import (CollectionSerializer, ProcessSerializer, DataSerializer,
                          DescriptorSchemaSerializer, TriggerSerializer, StorageSerializer)

from resolwe.permissions.shortcuts import get_user_group_perms


def assign_perm(*args, **kwargs):
    """Wrapper for assign_perm function

    Call original assign_perms function from django-guardian, but don't
    raise exception if permission is not found

    """
    try:
        shortcuts.assign_perm(*args, **kwargs)
    except Permission.DoesNotExist:
        pass


def remove_perm(*args, **kwargs):
    """Wrapper for remove_perm function

    Call original remove_perms function from django-guardian, but don't
    raise exception if permission is not found

    """
    try:
        shortcuts.remove_perm(*args, **kwargs)
    except Permission.DoesNotExist:
        pass


def load_permissions(permissions_name):
    """Look for a fully qualified flow permissions class."""
    try:
        return import_module('{}'.format(permissions_name)).ResolwePermissions
    except AttributeError:
        raise AttributeError("'ResolwePermissions' class not found in {} module.".format(
                             permissions_name))
    except ImportError as ex:
        # The permissions module wasn't found. Display a helpful error
        # message listing all possible (built-in) permissions classes.
        permissions_dir = os.path.join(os.path.dirname(upath(__file__)), '..', 'perms')
        permissions_dir = os.path.normpath(permissions_dir)

        try:
            builtin_permissions = [
                name for _, name, _ in pkgutil.iter_modules([permissions_dir]) if name not in [u'tests']]
        except EnvironmentError:
            builtin_permissions = []
        if permissions_name not in ['resolwe.auth.{}'.format(p) for p in builtin_permissions]:
            permissions_reprs = map(repr, sorted(builtin_permissions))
            err_msg = ("{} isn't an available flow permissions class.\n"
                       "Try using 'resolwe.auth.XXX', where XXX is one of:\n"
                       "    {}\n"
                       "Error was: {}".format(permissions_name, ", ".join(permissions_reprs), ex))
            raise ImproperlyConfigured(err_msg)
        else:
            # If there's some other error, this must be an error in Django
            raise


permissions_cls = load_permissions(settings.FLOW_API['PERMISSIONS'])


class ResolweCreateModelMixin(mixins.CreateModelMixin):
    """Mixin to support creating new `Resolwe` models

    Extends `django_rest_framework`'s class `CreateModelMixin` with:

      * append user's id from request to posted data as `contributor`
        key
      * catch `IntegrityError`s, so we can return HTTP status 409
        instead of raising error

    """
    def create(self, request, *args, **kwargs):
        user = request.user
        if not user.is_authenticated():
            raise exceptions.NotFound

        ds_slug = request.data.get('descriptor_schema', None)
        ds_query = DescriptorSchema.objects.filter(slug=ds_slug).order_by('version')
        if ds_slug and not ds_query.exists():
            return Response(
                {'descriptor_schema':
                    ['Invalid descriptor_schema slug "{}" - object does not exist.'.format(ds_slug)]},
                status=status.HTTP_400_BAD_REQUEST)

        request.data['contributor'] = user.pk
        try:
            return super(ResolweCreateModelMixin, self).create(request, *args, **kwargs)
        except IntegrityError as ex:
            return Response({u'error': str(ex)}, status=status.HTTP_409_CONFLICT)

    def perform_create(self, serializer):
        with transaction.atomic():
            instance = serializer.save()

            # Assign all permissions to the object contributor.
            for permission in list(zip(*instance._meta.permissions))[0]:
                assign_perm(permission, instance.contributor, instance)

        ds_slug = self.request.data.get('descriptor_schema', None)
        if ds_slug:
            descriptor_schema = DescriptorSchema.objects.filter(slug=ds_slug).order_by('version').last()
            instance.descriptor_schema = descriptor_schema
            instance.save()


class ResolweCreateDataModelMixin(ResolweCreateModelMixin):
    """Mixin to support creating new :class:`Data` objects

    Extends :class:`ResolweCcreateModelMixin` with:

      * checks if there is exactly 1 collection listed on create
      * checks if user has `add` permission on that collection

    """
    def create(self, request, *args, **kwargs):
        collections = request.data.get('collections', [])

        for collection_id in collections:
            try:
                collection = Collection.objects.get(pk=collection_id)
            except Collection.DoesNotExist:
                return Response({'collections': ['Invalid pk "{}" - object does not exist.'.format(collection_id)]},
                                status=status.HTTP_400_BAD_REQUEST)

            if not request.user.has_perm('add_collection', obj=collection):
                if request.user.is_authenticated():
                    raise exceptions.PermissionDenied
                else:
                    raise exceptions.NotFound

        process_slug = request.data.get('process', None)
        process_query = Process.objects.filter(slug=process_slug).order_by('version')
        if not process_query.exists():
            # XXX: security - is it ok to reveal which processes (don't) exist?
            return Response({'process': ['Invalid process slug "{}" - object does not exist.'.format(process_slug)]},
                            status=status.HTTP_400_BAD_REQUEST)
        process = process_query.last()
        request.data['process'] = process.pk

        ds_slug = request.data.get('descriptor_schema', None)
        ds_query = DescriptorSchema.objects.filter(slug=ds_slug).order_by('version')
        if ds_slug and not ds_query.exists():
            return Response(
                {'descriptor_schema':
                    ['Invalid descriptor_schema slug "{}" - object does not exist.'.format(ds_slug)]},
                status=status.HTTP_400_BAD_REQUEST)

        if not request.user.has_perm('view_process', obj=process):
            if request.user.is_authenticated():
                raise exceptions.PermissionDenied
            else:
                raise exceptions.NotFound

        resp = super(ResolweCreateDataModelMixin, self).create(request, *args, **kwargs)
        manager.communicate()
        return resp

    def perform_create(self, serializer):
        with transaction.atomic():
            instance = serializer.save()

            # Assign all permissions to the object contributor.
            for permission in list(zip(*instance._meta.permissions))[0]:
                assign_perm(permission, instance.contributor, instance)

        collections = self.request.data.get('collections', [])
        for c in collections:
            collection = Collection.objects.get(pk=c)
            collection.data.add(instance)

        ds_slug = self.request.data.get('descriptor_schema', None)
        if ds_slug:
            descriptor_schema = DescriptorSchema.objects.filter(slug=ds_slug).order_by('version').last()
            instance.descriptor_schema = descriptor_schema
            instance.save()


class ResolwePermissionsMixin(object):
    """Mixin to support managing `Resolwe` objects' permissions.



    """
    def filter_public_permisions(self, perms):
        """Return list of parameters applicable for public user.

        :param list perms: List of permissions to filter
        :return: List of parameters
        :rtype: list

        """
        return [perm for perm in perms if perm.startswith('view') or perm.startswith('download')]

    def _fetch_user(self, query):
        try:
            return get_user_model().objects.get(Q(pk=query) | Q(email=query))
        except get_user_model().DoesNotExist:
            return None

    def _fetch_group(self, query):
        try:
            return Group.objects.get(Q(pk=query) | Q(name=query))
        except Group.DoesNotExist:
            return None

    def _update_permission(self, obj, data):
        content_type = ContentType.objects.get_for_model(obj)
        full_permissions = list(zip(*obj._meta.permissions))[0]

        def set_permissions(entity_type, perm_type):
            perm_func = assign_perm if perm_type == 'add' else remove_perm
            fetch = self._fetch_user if entity_type == 'users' else self._fetch_group

            for entity_id in data.get(entity_type, {}).get(perm_type, []):
                entity = fetch(entity_id)
                if entity:
                    perms = data[entity_type][perm_type][entity_id]
                    if perms == u'ALL':
                        perms = full_permissions
                    for perm in perms:
                        perm_func('{}_{}'.format(perm.lower(), content_type), entity, obj)

        set_permissions('users', 'add')
        set_permissions('users', 'remove')
        set_permissions('groups', 'add')
        set_permissions('groups', 'remove')

        def set_public_permissions(perm_type):
            perm_func = assign_perm if perm_type == 'add' else remove_perm
            user = AnonymousUser()
            perms = data.get('public', {}).get(perm_type, [])
            if perms == u'ALL':
                perms = full_permissions
            perms = self.filter_public_permisions(perms)
            for perm in perms:
                perm_func('{}_{}'.format(perm.lower(), content_type), user, obj)

        set_public_permissions('add')
        set_public_permissions('remove')

    def _get_object_permissions(self, user, instance):
        def format_permissions(perms):
            return [perm.replace('_{}'.format(instance._meta.model_name), '') for perm in perms]

        if user.is_authenticated():
            permissions_user, permissions_group = get_user_group_perms(user, instance)
        else:
            permissions_user, permissions_group = [], []
        permissions_public = shortcuts.get_perms(AnonymousUser(), instance)

        return {
            'user': format_permissions(permissions_user),
            'group': format_permissions(permissions_group),
            'public': format_permissions(permissions_public),
        }

    def get_serializer_class(self):
        # Augment the base serializer class to include permissions information with objects.
        base_class = super(ResolwePermissionsMixin, self).get_serializer_class()

        class SerializerWithPermissions(base_class):
            def to_representation(serializer, instance):
                # TODO: These permissions queries may be expensive. Should we limit or optimize this?
                data = super(SerializerWithPermissions, serializer).to_representation(instance)
                data['permissions'] = self._get_object_permissions(self.request.user, instance)
                return data

        return SerializerWithPermissions

    @detail_route(methods=[u'post'], url_path='permissions')
    def detail_permissions(self, request, pk=None):
        obj = self.get_object()
        content_type = ContentType.objects.get_for_model(obj)

        if not request.user.has_perm('share_{}'.format(content_type), obj=obj):
            if request.user.is_authenticated():
                raise exceptions.PermissionDenied()
            else:
                raise exceptions.NotFound()

        self._update_permission(obj, request.data)

        # resp = UserObjectPermission.objects.filter(object_pk=obj.pk)
        return Response()

    @list_route(methods=[u'post'], url_path='permissions')
    def list_permissions(self, request):
        # TODO
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)


class ResolweProcessPermissionsMixin(ResolwePermissionsMixin):

    def _update_permission(self, obj, data):
        super(ResolweProcessPermissionsMixin, self)._update_permission(obj, data)

        if 'collections' in data:
            if 'add' in data['collections']:
                for _id in data['collections']['add']:
                    try:
                        Collection.objects.get(pk=_id).public_processes.add(obj)
                        # obj.collections.add(Collection.objects.get(pk=_id))
                    except Collection.DoesNotExist:
                        pass
            if 'remove' in data['collections']:
                for _id in data['collections']['remove']:
                    try:
                        Collection.objects.get(pk=_id).public_processes.remove(obj)
                        # obj.collections.remove(Collection.objects.get(pk=_id))
                    except Collection.DoesNotExist:
                        pass


class ResolweCheckSlugMixin(object):
    @list_route(methods=[u'get'])
    def slug_exists(self, request):
        """Check if given url slug exists.

        Check if slug given in query parameter ``name`` exists. Return
        ``True`` if slug already exists and ``False`` otherwise.

        """
        if not request.user.is_authenticated():
            return Response(status=status.HTTP_401_UNAUTHORIZED)

        if 'name' not in request.query_params:
            return Response({'error': 'Query parameter `name` must be given.'},
                            status=status.HTTP_400_BAD_REQUEST)

        queryset = self.get_queryset()
        slug_name = request.query_params['name']
        return Response(queryset.filter(slug__iexact=slug_name).exists())


class DescriptorSchemaFilter(filters.FilterSet):
    class Meta:
        model = DescriptorSchema
        fields = {
            'slug': filters.ALL_LOOKUPS,
        }


class CollectionFilter(filters.FilterSet):
    data = filters.ModelChoiceFilter(queryset=Data.objects.all())
    descriptor_schema = filters.RelatedFilter(DescriptorSchemaFilter, name='descriptor_schema')

    class Meta:
        model = Collection
        fields = ('contributor', 'name', 'description', 'created', 'modified',
                  'slug', 'descriptor', 'data', 'descriptor_schema', 'id')


class CollectionViewSet(ResolweCreateModelMixin,
                        mixins.RetrieveModelMixin,
                        mixins.UpdateModelMixin,
                        mixins.DestroyModelMixin,
                        mixins.ListModelMixin,
                        ResolwePermissionsMixin,
                        ResolweCheckSlugMixin,
                        viewsets.GenericViewSet):

    """API view for :class:`Collection` objects."""

    queryset = Collection.objects.all().prefetch_related('descriptor_schema')
    serializer_class = CollectionSerializer
    permission_classes = (permissions_cls,)
    filter_class = CollectionFilter

    @detail_route(methods=[u'post'])
    def add_data(self, request, pk=None):
        collection = self.get_object()

        if not request.user.has_perm('add_collection', obj=collection):
            if request.user.is_authenticated():
                raise exceptions.PermissionDenied()
            else:
                raise exceptions.NotFound()

        if 'ids' not in request.data:
            return Response({"error": "`ids`parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        missing = []
        for data_id in request.data['ids']:
            if not Data.objects.filter(pk=data_id).exists():
                missing.append(data_id)

        if missing:
            return Response(
                {"error": "Data objects with following ids are missing: {}".format(', '.join(missing))},
                status=status.HTTP_400_BAD_REQUEST)

        for data_id in request.data['ids']:
            collection.data.add(data_id)

        return Response()

    @detail_route(methods=[u'post'])
    def remove_data(self, request, pk=None):
        collection = self.get_object()

        if not request.user.has_perm('add_collection', obj=collection):
            if request.user.is_authenticated():
                raise exceptions.PermissionDenied()
            else:
                raise exceptions.NotFound()

        if 'ids' not in request.data:
            return Response({"error": "`ids`parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        for data_id in request.data['ids']:
            collection.data.remove(data_id)

        return Response()


class ProcessFilter(filters.FilterSet):
    category = filters.CharFilter(name='category', lookup_type='startswith')

    class Meta:
        model = Process
        fields = ('contributor', 'name', 'created', 'modified', 'slug', 'category')


class ProcessViewSet(mixins.RetrieveModelMixin,
                     mixins.ListModelMixin,
                     ResolweProcessPermissionsMixin,
                     viewsets.GenericViewSet):

    """API view for :class:`Process` objects."""

    queryset = Process.objects.all()
    serializer_class = ProcessSerializer
    permission_classes = (permissions_cls,)
    filter_class = ProcessFilter


class DataFilter(filters.FilterSet):
    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    type = filters.CharFilter(name='process__type', lookup_type='startswith')

    class Meta:
        model = Data
        fields = ('contributor', 'name', 'created', 'modified', 'slug', 'input', 'descriptor',
                  'started', 'finished', 'output', 'status', 'process', 'type', 'collection')


class DataViewSet(ResolweCreateDataModelMixin,
                  mixins.RetrieveModelMixin,
                  mixins.UpdateModelMixin,
                  mixins.DestroyModelMixin,
                  mixins.ListModelMixin,
                  ResolwePermissionsMixin,
                  ResolweCheckSlugMixin,
                  viewsets.GenericViewSet):

    """API view for :class:`Data` objects."""

    queryset = Data.objects.all().prefetch_related('process', 'descriptor_schema')
    serializer_class = DataSerializer
    permission_classes = (permissions_cls,)
    filter_class = DataFilter


class DescriptorSchemaViewSet(mixins.RetrieveModelMixin,
                              mixins.ListModelMixin,
                              ResolwePermissionsMixin,
                              viewsets.GenericViewSet):

    """API view for :class:`DescriptorSchema` objects."""

    queryset = DescriptorSchema.objects.all()
    serializer_class = DescriptorSchemaSerializer
    permission_classes = (permissions_cls,)
    filter_fields = ('contributor', 'name', 'description', 'created', 'modified', 'slug')


class TriggerViewSet(ResolweCreateModelMixin,
                     mixins.RetrieveModelMixin,
                     mixins.UpdateModelMixin,
                     mixins.DestroyModelMixin,
                     mixins.ListModelMixin,
                     ResolwePermissionsMixin,
                     viewsets.GenericViewSet):

    """API view for :class:`Trigger` objects."""

    queryset = Trigger.objects.all()
    serializer_class = TriggerSerializer
    permission_classes = (permissions_cls,)
    filter_fields = ('contributor', 'name', 'created', 'modified', 'slug', 'collection')


class StorageViewSet(mixins.RetrieveModelMixin,
                     mixins.ListModelMixin,
                     viewsets.GenericViewSet):

    """API view for :class:`Storage` objects."""

    queryset = Storage.objects.all()
    serializer_class = StorageSerializer
    filter_fields = ('contributor', 'name', 'created', 'modified', 'slug', 'json')



"""
================
Resource Filters
================

"""
import os
from django import template

register = template.Library()  # pylint: disable=invalid-name


@register.filter
def basename(path):
    """Return the base name of pathname path."""
    return os.path.basename(path)


@register.filter
def subtype(type_, supertype):
    """Check if type_ is a subtype of supertype."""
    return type_.startswith(supertype)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import mock
import unittest

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse

from rest_framework.test import APIRequestFactory, force_authenticate
from rest_framework import status

from resolwe.flow.models import Data, Process, Collection
from resolwe.flow.views import DataViewSet, CollectionViewSet

factory = APIRequestFactory()


MESSAGES = {
    u'NOT_FOUND': u'Not found.',
}

class TestDataViewSetCase(unittest.TestCase):
    def setUp(self):
        self.data_viewset = DataViewSet.as_view(actions={
            'get': 'list',
        })

        self.user = User.objects.create(is_superuser=True)

    def tearDown(self):
        Data.objects.all().delete()
        Process.objects.all().delete()
        self.user.delete()

    @mock.patch('resolwe.flow.models.Process.objects.all')
    def test_prefetch(self, process_mock):
        # TODO: find way to mock Data objects
        proc = Process.objects.create(type='test:process', name='Test process', contributor=self.user)
        Data.objects.create(contributor=self.user, slug='test1', process=proc)
        Data.objects.create(contributor=self.user, slug='test2', process=proc)

        request = factory.get('/', '', content_type='application/json')
        force_authenticate(request, self.user)
        self.data_viewset(request)

        # check that only one request is made to get all processes' types
        self.assertEqual(process_mock.call_count, 1)


class TestCollectionViewSetCase(unittest.TestCase):
    def setUp(self):
        self.checkslug_viewset = CollectionViewSet.as_view(actions={
            'get': 'slug_exists',
        })
        self.add_data_viewset = CollectionViewSet.as_view(actions={
            'post': 'add_data',
        })
        self.remove_data_viewset = CollectionViewSet.as_view(actions={
            'post': 'remove_data',
        })

        self.detail_url = lambda pk: reverse('resolwe-api:collection-detail', kwargs={'pk': pk})

        self.super_user = User.objects.create(username='superuser', is_superuser=True)
        self.user = User.objects.create(username='normaluser')

    def tearDown(self):
        Data.objects.all().delete()
        Process.objects.all().delete()
        Collection.objects.all().delete()
        self.super_user.delete()
        self.user.delete()

    def test_check_slug(self):
        Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.super_user)

        # unauthorized
        request = factory.get('/', {'name': 'collection1'}, content_type='application/json')
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.status_code, 401)
        self.assertEqual(resp.data, None)

        # existing slug
        request = factory.get('/', {'name': 'collection1'}, content_type='application/json')
        force_authenticate(request, self.super_user)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, True)

        # existing slug - iexact
        request = factory.get('/', {'name': 'Collection1'}, content_type='application/json')
        force_authenticate(request, self.super_user)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, True)

        # non-existing slug
        request = factory.get('/', {'name': 'new-collection'}, content_type='application/json')
        force_authenticate(request, self.super_user)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, False)

        # bad query parameter
        request = factory.get('/', {'bad': 'parameter'}, content_type='application/json')
        force_authenticate(request, self.super_user)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.status_code, 400)

    def test_add_remove_data(self):
        c = Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.super_user)

        proc = Process.objects.create(type='test:process', name='Test process', contributor=self.user)
        d = Data.objects.create(contributor=self.user, slug='test1', process=proc)

        request = factory.post(self.detail_url(c.pk), '{{"ids": ["{}"]}}'.format(d.pk),
                               content_type='application/json')

        # user w/o permissions cannot add data
        force_authenticate(request, self.user)
        resp = self.add_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.data[u'detail'], MESSAGES['NOT_FOUND'])
        self.assertEqual(c.data.count(), 0)

        # user w/ permissions can add data
        force_authenticate(request, self.super_user)
        resp = self.add_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(c.data.count(), 1)

        request = factory.post(self.detail_url(c.pk), '{{"ids": ["{}"]}}'.format(d.pk),
                               content_type='application/json')

        # user w/o permissions cannot add data
        force_authenticate(request, self.user)
        resp = self.remove_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.data[u'detail'], MESSAGES['NOT_FOUND'])
        self.assertEqual(c.data.count(), 1)

        # user w/ permissions can remove data
        force_authenticate(request, self.super_user)
        resp = self.remove_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(c.data.count(), 0)


        request = factory.post(self.detail_url(c.pk), '{"ids": ["42"]}', content_type='application/json')

        force_authenticate(request, self.super_user)
        resp = self.remove_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(c.data.count(), 0)



from __future__ import absolute_import, division, print_function, unicode_literals

import sys

try:
    import celery
except ImportError:
    print('Please install Celery using `pip install celery`', file=sys.stderr)
    sys.exit(1)


from .base import BaseManager
from ..tasks import celery_run


class Manager(BaseManager):

    def run(self, data_id, script, run_sync=False, verbosity=1):
        celery_run.delay(self.executor, data_id, script, verbosity)



"""Workflow workload managers"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import pkgutil

from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils._os import upath


__all__ = ['manager']


def load_manager(manager_name):
    try:
        return import_module('{}'.format(manager_name))
    except ImportError as ex:
        # The manager wasn't found. Display a helpful error message
        # listing all possible (built-in) managers.
        manager_dir = os.path.join(os.path.dirname(upath(__file__)), 'managers')

        try:
            builtin_managers = [name for _, name, _ in pkgutil.iter_modules([manager_dir])]
        except EnvironmentError:
            builtin_managers = []
        if manager_name not in ['resolwe.flow.managers.{}'.format(b) for b in builtin_managers]:
            manager_reprs = map(repr, sorted(builtin_managers))
            error_msg = ("{} isn't an available managers.\n"
                         "Try using 'resolwe.flow.managers.XXX', where XXX is one of:\n"
                         "    {}\n"
                         "Error was: {}".format(manager_name, ", ".join(manager_reprs), ex))
            raise ImproperlyConfigured(error_msg)
        else:
            # If there's some other error, this must be an error in Django
            raise


settings_flow_manager = getattr(settings, 'FLOW_MANAGER', 'resolwe.flow.managers.local')
manager = load_manager(settings_flow_manager).Manager()  # pylint: disable=invalid-name



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import mock
import shlex
import subprocess
import unittest

from django.conf import settings
from django.test import override_settings

from ..executors.docker import FlowExecutor

try:
    import builtins  # py3
except ImportError:
    import __builtin__ as builtins  # py2


def check_docker():
    """Check if Docker is installed and working.

    :return: tuple (indicator of the availability of Docker, reason for
             unavailability)
    :rtype: (bool, str)

    """
    command = settings.FLOW_EXECUTOR.get('COMMAND', 'docker')
    info_command = '{} info'.format(command)
    available, reason = True, ""
    try:
        subprocess.check_call(shlex.split(info_command), stdout=subprocess.PIPE)
    except OSError:
        available, reason = False, "Docker command '{}' not found".format(command)
    except subprocess.CalledProcessError:
        available, reason = (False, "Docker command '{}' returned non-zero "
                                    "exit status".format(info_command))
    return available, reason


class DockerExecutorTestCase(unittest.TestCase):
    @unittest.skipUnless(*check_docker())
    @mock.patch('os.mkdir')
    @mock.patch('os.chmod')
    @mock.patch('os.chdir')
    @mock.patch('resolwe.flow.executors.Data.objects.filter')
    @mock.patch('resolwe.flow.executors.Data.objects.get')
    def test_run_in_docker(self, data_get_mock, data_filter_mock, chdir_mock, chmod_mock, mkdir_mock):
        executor_settings = settings.FLOW_EXECUTOR
        executor_settings['CONTAINER_IMAGE'] = 'centos'

        with override_settings(FLOW_EXECUTOR=executor_settings):
            executor = FlowExecutor()

            script = 'if [[ -f /.dockerinit ]]; then echo "Running inside Docker"; else echo "Running locally"; fi'

            count = {'running': 0}

            def assert_output(line):
                if line.strip() == 'Running inside Docker':
                    count['running'] += 1

            write_mock = mock.MagicMock(side_effect=assert_output)
            stdout_mock = mock.MagicMock(write=write_mock)
            open_mock = mock.MagicMock(side_effect=[stdout_mock, mock.MagicMock()])
            with mock.patch.object(builtins, 'open', open_mock):
                executor.run('no_data_id', script, verbosity=0)

            self.assertEqual(count['running'], 1)



from __future__ import absolute_import, division, print_function, unicode_literals

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse

from rest_framework.test import APITestCase, APIRequestFactory, force_authenticate


class ResolweAPITestCase(APITestCase):
    def setUp(self):
        super(ResolweAPITestCase, self).setUp()

        self.factory = APIRequestFactory()

        self.user1 = User.objects.get(pk=1)
        self.user2 = User.objects.get(pk=2)
        self.user3 = User.objects.get(pk=3)
        self.admin = User.objects.get(pk=4)

        list_url_mapping = {}
        if hasattr(self.viewset, 'list'):
            list_url_mapping['get'] = 'list'
        if hasattr(self.viewset, 'create'):
            list_url_mapping['post'] = 'create'

        self.list_view = self.viewset.as_view(list_url_mapping)

        detail_url_mapping = {}
        if hasattr(self.viewset, 'retrieve'):
            detail_url_mapping['get'] = 'retrieve'
        if hasattr(self.viewset, 'update'):
            detail_url_mapping['put'] = 'update'
        if hasattr(self.viewset, 'partial_update'):
            detail_url_mapping['patch'] = 'partial_update'
        if hasattr(self.viewset, 'destroy'):
            detail_url_mapping['delete'] = 'destroy'
        if hasattr(self.viewset, 'detail_permissions'):
            detail_url_mapping['post'] = 'detail_permissions'

        self.detail_view = self.viewset.as_view(detail_url_mapping)

        self.detail_url = lambda pk: reverse('resolwe-api:{}-detail'.format(self.resource_name), kwargs={'pk': pk})
        self.detail_permissions = lambda pk: reverse('resolwe-api:{}-permissions'.format(self.resource_name),
                                                     kwargs={'pk': pk})
        self.list_url = reverse('resolwe-api:collection-list')

    def _get_list(self, user=None):
        request = self.factory.get(self.list_url, format='json')
        force_authenticate(request, user)
        resp = self.list_view(request)
        resp.render()
        return resp

    def _get_detail(self, pk, user=None):
        request = self.factory.get(self.detail_url(pk), format='json')
        force_authenticate(request, user)
        resp = self.detail_view(request, pk=pk)
        resp.render()
        return resp

    def _post(self, data={}, user=None):
        request = self.factory.post(self.list_url, data=data, format='json')
        force_authenticate(request, user)
        resp = self.list_view(request)
        resp.render()
        return resp

    def _patch(self, pk, data={}, user=None):
        request = self.factory.patch(self.detail_url(pk), data=data, format='json')
        force_authenticate(request, user)
        resp = self.detail_view(request, pk=pk)
        resp.render()
        return resp

    def _delete(self, pk, user=None):
        request = self.factory.delete(self.detail_url(pk), format='json')
        force_authenticate(request, user)
        resp = self.detail_view(request, pk=pk)
        resp.render()
        return resp

    def _detail_permissions(self, pk, data={}, user=None):
        request = self.factory.post(self.detail_permissions(pk), data=data, format='json')
        force_authenticate(request, user)
        resp = self.detail_view(request, pk=pk)
        resp.render()
        return resp

    def assertKeys(self, data, wanted):
        self.assertEqual(sorted(data.keys()), sorted(wanted))



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil

from datetime import datetime, timedelta


from django.conf import settings
if settings.USE_TZ:
    from django.utils.timezone import now
else:
    now = datetime.now  # pylint: disable=invalid-name

from rest_framework import status

from .base import ResolweAPITestCase
from resolwe.flow.models import Data, Collection
from resolwe.flow.views import DataViewSet


DATE_FORMAT = r'%Y-%m-%dT%H:%M:%S.%f'

MESSAGES = {
    u'NOT_FOUND': u'Not found.',
    # u'NO_PERMISSION': u'You do not have permission to perform this action.',
    u'ONE_ID_REQUIRED': 'Exactly one id required on create.',
}


class DataTestCase(ResolweAPITestCase):
    fixtures = ['users.yaml', 'collections.yaml', 'permissions.yaml', 'processes.yaml', 'data.yaml']

    def setUp(self):
        # self.data1 = Data.objects.get(pk=1)

        self.resource_name = 'data'
        self.viewset = DataViewSet

        self.data = {
            'name': 'New data',
            'slug': 'new_data',
            'collections': ['1'],
            'process': 'test_process',
        }

        super(DataTestCase, self).setUp()

    def tearDown(self):
        for data in Data.objects.all():
            data_dir = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(data.id))
            shutil.rmtree(data_dir, ignore_errors=True)


    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_get_list_public_user(self):
        resp = self._get_list()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 1)

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_post(self):
        # logged-in user w/ perms
        collection_n = Data.objects.count()
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Data.objects.count(), collection_n + 1)

        d = Data.objects.get(pk=resp.data['id'])
        self.assertTrue(now() - d.modified < timedelta(seconds=1))
        self.assertTrue(now() - d.created < timedelta(seconds=1))
        self.assertEqual(d.status, 'OK')
        self.assertTrue(now() - d.started < timedelta(seconds=1))
        self.assertTrue(now() - d.finished < timedelta(seconds=1))
        self.assertEqual(d.contributor_id, 1)

    def test_post_invalid_fields(self):
        data_n = Data.objects.count()

        self.data['collections'] = ['42']
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data[u'collections'][0], u'Invalid pk "42" - object does not exist.')

        self.data['collections'] = ['1']
        self.data['process'] = '42'
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data[u'process'][0], u'Invalid process slug "42" - object does not exist.')

        self.assertEqual(Data.objects.count(), data_n)

    def test_post_no_perms(self):
        collection_n = Data.objects.count()
        resp = self._post(self.data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(Data.objects.count(), collection_n)

    def test_post_public_user(self):
        collection_n = Data.objects.count()
        resp = self._post(self.data)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(Data.objects.count(), collection_n)

    def test_post_protected_fields(self):
        date_now = now()
        self.data['created'] = date_now - timedelta(days=360)
        self.data['modified'] = date_now - timedelta(days=180)
        self.data['started'] = date_now - timedelta(days=180)
        self.data['finished'] = date_now - timedelta(days=90)
        self.data['checksum'] = 'fake'
        self.data['status'] = 'DE'
        self.data['process_progress'] = 2
        self.data['process_rc'] = 18
        self.data['process_info'] = 'Spam'
        self.data['process_warning'] = 'More spam'
        self.data['process_error'] = 'Even more spam'
        self.data['contributor_id'] = 2

        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.assertEqual(resp.data['started'], None)
        self.assertEqual(resp.data['finished'], None)
        self.assertEqual(resp.data['checksum'], None)  # TODO: Add checksum when implemented
        self.assertEqual(resp.data['status'], 'RE')
        self.assertEqual(resp.data['process_progress'], 0)
        self.assertEqual(resp.data['process_rc'], None)
        self.assertEqual(resp.data['process_info'], [])
        self.assertEqual(resp.data['process_warning'], [])
        self.assertEqual(resp.data['process_error'], [])
        self.assertEqual(resp.data['contributor'], 1)

    def test_post_multiple_collections(self):
        self.data['collections'].append('2')

        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Collection.objects.filter(data=resp.data['id']).count(), 2)

    def test_get_detail(self):
        # public user w/ perms
        resp = self._get_detail(1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data[u'id'], 1)

        # user w/ permissions
        resp = self._get_detail(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(resp.data, [u'slug', u'name', u'created', u'modified', u'contributor',
                                    u'started', u'finished', u'checksum', u'status', u'process',
                                    u'process_progress', u'process_rc', u'process_info', u'process_warning',
                                    u'process_error', u'input', u'output', u'process_type', u'descriptor_schema',
                                    u'descriptor', u'id', u'process_name', u'process_input_schema',
                                    u'process_output_schema', u'permissions'])

    def test_get_detail_no_perms(self):
        # public user w/o permissions
        resp = self._get_detail(2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data[u'detail'], MESSAGES['NOT_FOUND'])

        # user w/o permissions
        resp = self._get_detail(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data[u'detail'], MESSAGES['NOT_FOUND'])

    def test_patch(self):
        data = {'name': 'New data'}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.name, 'New data')

    def test_patch_no_perms(self):
        data = {'name': 'New data'}
        resp = self._patch(2, data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        d = Data.objects.get(pk=2)
        self.assertEqual(d.name, 'Test data 2')

    def test_patch_public_user(self):
        data = {'name': 'New data'}
        resp = self._patch(2, data)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        d = Data.objects.get(pk=2)
        self.assertEqual(d.name, 'Test data 2')

    def test_patch_protected(self):
        date_now = now()

        # `created`
        resp = self._patch(1, {'created': date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.created.isoformat(), datetime(2015, 1, 1, 9, 0, 0).isoformat())
        self.assertEqual(d.modified.isoformat(), datetime(2015, 1, 1, 9, 0, 0).isoformat())

        # `modified`
        resp = self._patch(1, {'modified': date_now - timedelta(days=180)}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `started`
        resp = self._patch(1, {'started': date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.started, datetime(2015, 1, 1, 9, 0, 0))
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `finished`
        resp = self._patch(1, {'finished': date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.finished, datetime(2015, 1, 1, 9, 0, 0))
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `checksum`
        resp = self._patch(1, {'checksum': 'fake'}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.checksum, None)  # TODO: Add checksum when implemented
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `status`
        resp = self._patch(1, {'status': 'DE'}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.status, 'OK')
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `process_progress`
        resp = self._patch(1, {'process_progress': 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_progress, 0)
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `process_rc`
        resp = self._patch(1, {'process_rc': 18}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_rc, None)
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `process_info`
        resp = self._patch(1, {'process_info': 'Spam'}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_info, [])
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `process_warning`
        resp = self._patch(1, {'process_warning': 'More spam'}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_warning, [])
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `process_error`
        resp = self._patch(1, {'process_error': 'Even more spam'}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_error, [])
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `contributor`
        resp = self._patch(1, {'contributor': 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.contributor_id, 1)
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

        # `process`
        resp = self._patch(1, {'process': 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_id, 1)
        self.assertEqual(d.modified, datetime(2015, 1, 1, 9, 0, 0))

    def test_delete(self):
        resp = self._delete(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        q = Data.objects.filter(pk=1).exists()
        self.assertFalse(q)

    def test_delete_no_perms(self):
        resp = self._delete(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        q = Data.objects.filter(pk=2).exists()
        self.assertTrue(q)

    def test_delete_public_user(self):
        resp = self._delete(2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        q = Data.objects.filter(pk=2).exists()
        self.assertTrue(q)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import pkgutil

from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import IntegrityError, transaction
from django.utils._os import upath

from resolwe.flow.models import Data, iterate_fields
from resolwe.utils import BraceMessage as __


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def dependency_status(data):
    """Return abstracted satus of dependencies.

    STATUS_ERROR .. one dependency has error status
    STATUS_DONE .. all dependencies have done status
    None .. other

    """
    for field_schema, fields in iterate_fields(data.input, data.process.input_schema):
        if (field_schema['type'].lower().startswith('data:') or
                field_schema['type'].lower().startswith('list:data:')):
            name = field_schema['name']
            value = fields[name]

            if field_schema['type'].lower().startswith('data:'):
                value = [value]

            for uid in value:
                try:
                    _data = Data.objects.get(id=uid)
                except Data.DoesNotExist:
                    return Data.STATUS_ERROR

                if _data.status == Data.STATUS_ERROR:
                    return Data.STATUS_ERROR

                if _data.status != Data.STATUS_DONE:
                    return None

    return Data.STATUS_DONE


class BaseManager(object):

    """Manager handles process job execution."""

    def __init__(self):
        self.executor = self.load_executor(settings.FLOW_EXECUTOR['NAME']).FlowExecutor()
        self.exprengines = self.load_exprengines(settings.FLOW_EXPRESSION_ENGINES)

    def run(self, data_id, script, run_sync=False, verbosity=1):
        raise NotImplementedError('`run` function not implemented')

    def communicate(self, run_sync=False, verbosity=1):
        """Resolving task dependancy and execution."""
        queue = []
        try:
            with transaction.atomic():
                for data in Data.objects.select_for_update().filter(status=Data.STATUS_RESOLVING):

                    dep_status = dependency_status(data)

                    if dep_status == Data.STATUS_ERROR:
                        data.status = Data.STATUS_ERROR
                        data.process_error.append("One or more inputs have status ERROR")
                        data.process_rc = 1
                        data.save()
                        continue

                    elif dep_status != Data.STATUS_DONE:
                        data.status = Data.STATUS_RESOLVING
                        data.save()
                        continue

                    script = self.exprengines['dtlbash'].eval(data)
                    if script is None:
                        continue

                    data.status = Data.STATUS_WAITING
                    data.save(render_name=True)

                    queue.append((data.id, script))

        except IntegrityError as exp:
            logger.error(__("IntegrityError in manager {}", exp))
            return

        for data_id, script in queue:
            if verbosity >= 1:
                print("Running", script)
            self.run(data_id, script, verbosity=verbosity)

    def load_executor(self, executor_name):
        try:
            return import_module('{}'.format(executor_name))
        except ImportError as ex:
            # The executor wasn't found. Display a helpful error message
            # listing all possible (built-in) executors.
            executor_dir = os.path.join(os.path.dirname(upath(__file__)), 'executors')

            try:
                builtin_executors = [name for _, name, _ in pkgutil.iter_modules([executor_dir])]
            except EnvironmentError:
                builtin_executors = []
            if executor_name not in ['resolwe.flow.executors.{}'.format(b) for b in builtin_executors]:
                executor_reprs = map(repr, sorted(builtin_executors))
                error_msg = ("{} isn't an available dataflow executors.\n"
                             "Try using 'resolwe.flow.executors.XXX', where XXX is one of:\n"
                             "    {}\n"
                             "Error was: {}".format(executor_name, ", ".join(executor_reprs), ex))
                raise ImproperlyConfigured(error_msg)
            else:
                # If there's some other error, this must be an error in Django
                raise

    def load_exprengines(self, exprengine_list):
        exprengines = {}

        for exprengine_name in exprengine_list:
            try:
                exprengine_module = import_module('{}'.format(exprengine_name))
                exprengine_key = exprengine_name.split('.')[-1]

                if exprengine_key in exprengines:
                    raise ImproperlyConfigured("Duplicated expression engine {}".format(exprengine_key))

                exprengines[exprengine_key] = exprengine_module.ExpressionEngine()

            except ImportError as ex:
                # The expression engine wasn't found. Display a helpful error message
                # listing all possible (built-in) expression engines.
                exprengine_dir = os.path.join(os.path.dirname(upath(__file__)), 'exprengines')

                try:
                    builtin_exprengines = [name for _, name, _ in pkgutil.iter_modules([exprengine_dir])]
                except EnvironmentError:
                    builtin_exprengines = []
                if exprengine_name not in ['resolwe.flow.exprengines.{}'.format(b) for b in builtin_exprengines]:
                    exprengine_reprs = map(repr, sorted(builtin_exprengines))
                    error_msg = ("{} isn't an available dataflow expression engine.\n"
                                 "Try using 'resolwe.flow.exprengines.XXX', where XXX is one of:\n"
                                 "    {}\n"
                                 "Error was: {}".format(exprengine_name, ", ".join(exprengine_reprs), ex))
                    raise ImproperlyConfigured(error_msg)
                else:
                    # If there's some other error, this must be an error in Django
                    raise

        return exprengines



"""Register processes"""
from __future__ import absolute_import, division, print_function, unicode_literals

import jsonschema
import os
import yaml

from django.core.management.base import BaseCommand
from django.contrib.auth import get_user_model
from django.db.models import Max
from django.utils.text import slugify

from versionfield.utils import convert_version_string_to_int

from resolwe.flow.models import DescriptorSchema, Process, iterate_schema, validation_schema, VERSION_NUMBER_BITS
from resolwe.flow.finders import get_finders


PROCESSOR_SCHEMA = validation_schema('processor')
DESCRIPTOR_SCHEMA = validation_schema('descriptor')


class Command(BaseCommand):

    """Register processes"""

    help = 'Register processes'

    def add_arguments(self, parser):
        parser.add_argument('-s', '--schemas', type=str, nargs='*', help="process names to register")
        parser.add_argument('-f', '--force', action='store_true', help="register also if version mismatch")

    def valid(self, instance, schema):
        """Validate schema."""
        try:
            jsonschema.validate(instance, schema)
            return True
        except jsonschema.exceptions.ValidationError as ex:
            self.stderr.write("    VALIDATION ERROR: {}".format(instance['name'] if 'name' in instance else ''))
            self.stderr.write("        path:       {}".format(ex.path))
            self.stderr.write("        message:    {}".format(ex.message))
            self.stderr.write("        validator:  {}".format(ex.validator))
            self.stderr.write("        val. value: {}".format(ex.validator_value))
            return False

    def find_schemas(self, schema_path, filters=None):
        """Find schemas in packages that match filters."""
        schema_matches = []

        if not os.path.isdir(schema_path):
            self.stdout.write("Invalid path {}".format(schema_path))
            return

        for root, _, files in os.walk(schema_path):
            for schema_file in [os.path.join(root, fn) for fn in files]:

                if not schema_file.lower().endswith(('.yml', '.yaml')):
                    continue

                schemas = yaml.load(open(schema_file))
                if not schemas:
                    self.stderr.write("Could not read YAML file {}".format(schema_file))
                    continue

                schema_matches.extend(schema for schema in schemas if
                                      not filters or schema.get('name', '') in filters or
                                      schema.get('slug', '') in filters)

        return schema_matches

    def register_processes(self, process_schemas, user, force=False):
        """Read and register processors."""
        log_processors = []
        log_templates = []

        for p in process_schemas:
            if p['type'][-1] != ':':
                p['type'] += ':'

            if 'category' in p and not p['category'].endswith(':'):
                p['category'] += ':'

            # get `data_name` from `static`
            if 'static' in p:
                for schema, _, _ in iterate_schema({}, p['static']):
                    if schema['name'] == 'name' and 'default' in schema:
                        p['data_name'] = schema['default']

            # support backward compatibility
            # TODO: update .yml files and remove
            if 'slug' not in p:
                p['slug'] = slugify(p.pop('name').replace(':', '-'))
                p['name'] = p.pop('label')

                p.pop('var', None)
                p.pop('static', None)

            for field in ['input', 'output', 'var', 'static']:
                for schema, _, _ in iterate_schema({}, p[field] if field in p else {}):
                    if not schema['type'][-1].endswith(':'):
                        schema['type'] += ':'
            # TODO: Check if schemas validate with our JSON meta schema and Processor model docs.

            if not self.valid(p, PROCESSOR_SCHEMA):
                continue

            if 'persistence' in p:
                persistence_mapping = {
                    'RAW': Process.PERSISTENCE_RAW,
                    'CACHED': Process.PERSISTENCE_CACHED,
                    'TEMP': Process.PERSISTENCE_TEMP,
                }

                p['persistence'] = persistence_mapping[p['persistence']]

            if 'input' in p:
                p['input_schema'] = p.pop('input')

            if 'output' in p:
                p['output_schema'] = p.pop('output')

            slug = p['slug']
            version = p['version']
            int_version = convert_version_string_to_int(version, VERSION_NUMBER_BITS)

            # `latest version` is returned as `int` so it has to be compared to `int_version`
            latest_version = Process.objects.filter(slug=slug).aggregate(Max('version'))['version__max']
            if latest_version is not None and latest_version > int_version:
                self.stderr.write("Skip processor {}: newer version installed".format(slug))
                continue

            process_query = Process.objects.filter(slug=slug, version=version)
            if process_query.exists():
                if not force:
                    self.stdout.write("Skip processor {}: same version installed".format(slug))
                    continue

                process_query.update(**p)
                log_processors.append("Updated {}".format(slug))
            else:
                Process.objects.create(contributor=user, **p)
                log_processors.append("Inserted {}".format(slug))

        if len(log_processors) > 0:
            self.stdout.write("Processor Updates:")
            for log in log_processors:
                self.stdout.write("  {}".format(log))

        if len(log_templates) > 0:
            self.stdout.write("Default Template Updates:")
            for log in log_templates:
                self.stdout.write("  {}".format(log))

    def register_descriptors(self, descriptor_schemas, user, force=False):
        """Read and register descriptors."""
        log_descriptors = []

        for ds in descriptor_schemas:

            for field in ['var', 'schema']:
                for schema, _, _ in iterate_schema({}, ds.get(field, {})):
                    if not schema['type'][-1].endswith(':'):
                        schema['type'] += ':'

            # support backward compatibility
            # TODO: update .yml files and remove
            if 'slug' not in ds:
                ds['slug'] = slugify(ds.pop('name').replace(':', '-'))
                ds['name'] = ds.pop('label')

            if 'schema' not in ds:
                ds['schema'] = []

            if 'static' in ds:
                ds['schema'].extend(ds.pop('static'))
            if 'var' in ds:
                ds['schema'].extend(ds.pop('var'))

            if not self.valid(ds, DESCRIPTOR_SCHEMA):
                continue

            slug = ds['slug']
            version = ds.get('version', '0.0.0')
            int_version = convert_version_string_to_int(version, VERSION_NUMBER_BITS)

            # `latest version` is returned as `int` so it has to be compared to `int_version`
            latest_version = DescriptorSchema.objects.filter(slug=slug).aggregate(Max('version'))['version__max']
            if latest_version is not None and latest_version > int_version:
                self.stderr.write("Skip descriptor schema {}: newer version installed".format(slug))
                continue

            descriptor_query = DescriptorSchema.objects.filter(slug=slug, version=version)
            if descriptor_query.exists():
                if not force:
                    self.stdout.write("Skip descriptor schema {}: same version installed".format(slug))
                    continue

                descriptor_query.update(**ds)
                log_descriptors.append("Updated {}".format(slug))
            else:
                DescriptorSchema.objects.create(contributor=user, **ds)
                log_descriptors.append("Inserted {}".format(slug))

        if len(log_descriptors) > 0:
            self.stdout.write("Descriptor schemas Updates:")
            for log in log_descriptors:
                self.stdout.write("  {}".format(log))

    def handle(self, *args, **options):
        schemas = options.get('schemas')
        force = options.get('force')

        users = get_user_model().objects.filter(is_superuser=True).order_by('date_joined')

        if not users.exists():
            self.stderr.write("Admin does not exist: create a superuser")
            exit(1)

        user_admin = users.first()

        processes_paths = []
        descriptors_paths = []
        for finder in get_finders():
            processes_paths.extend(finder.find_processes())
            descriptors_paths.extend(finder.find_descriptors())

        process_schemas = []
        for proc_path in processes_paths:
            process_schemas.extend(self.find_schemas(proc_path, schemas))

        self.register_processes(process_schemas, user_admin, force)

        descriptor_schemas = []
        for desc_path in descriptors_paths:
            descriptor_schemas.extend(self.find_schemas(desc_path))

        self.register_descriptors(descriptor_schemas, user_admin, force)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from rest_framework import status

from .base import ResolweAPITestCase
from resolwe.flow.models import Collection, Process
from resolwe.flow.views import ProcessViewSet


class ProcessTestCase(ResolweAPITestCase):
    fixtures = ['users.yaml', 'permissions.yaml', 'processes.yaml', 'collections.yaml']

    def setUp(self):
        self.process1 = Process.objects.get(pk=1)

        self.resource_name = 'process'
        self.viewset = ProcessViewSet

        super(ProcessTestCase, self).setUp()

    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_get_list_public_user(self):
        resp = self._get_list()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 1)

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_post(self):
        resp = self._post({'name': 'New process'}, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_patch(self):
        resp = self._patch(1, {'name': 'Hacked process'}, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete(self):
        resp = self._delete(1, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_collection_processes(self):
        data = {'collections': {'add': ['1']}}

        # add new `process` to `collection`
        self.assertFalse(Collection.objects.get(pk=1).public_processes.filter(pk=self.process1.pk).exists())
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, 200)
        self.process1.refresh_from_db()
        self.assertTrue(Collection.objects.get(pk=1).public_processes.filter(pk=self.process1.pk).exists())

        data = {'collections': {'remove': ['1']}}

        # remove existing `process` from `collection`
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, 200)
        self.process1.refresh_from_db()
        self.assertFalse(Collection.objects.get(pk=1).public_processes.filter(pk=self.process1.pk).exists())

        # remove non-existent `process` from `collection`
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, 200)

        data = {'collections': {'add': ['42', '1']}}

        # combination of valid and invalid `collection`s
        self.assertFalse(Collection.objects.get(pk=1).public_processes.filter(pk=self.process1.pk).exists())
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, 200)
        self.assertTrue(Collection.objects.get(pk=1).public_processes.filter(pk=self.process1.pk).exists())



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import mock
import unittest

from resolwe.flow.exprengines.dtlbash import ExpressionEngine


class ExprengineTestCase(unittest.TestCase):
    def test_invalid_template(self):
        data_mock = mock.MagicMock(process_error=[])
        data_mock.process.run = {'bash': '{% if reads.type.startswith("data:reads:") %}'}

        ExpressionEngine().eval(data_mock)

        self.assertTrue('Error in process script' in data_mock.process_error[0])


class CeleryEngineTestCase(unittest.TestCase):
    def test_passed_to_celery(self):
        pass



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime

from django.db import transaction

from rest_framework import status

from .base import ResolweAPITestCase
from resolwe.flow.models import Collection
from resolwe.flow.views import CollectionViewSet


MESSAGES = {
    u'NOT_FOUND': u'Not found.',
    u'NO_PERMISSION': u'You do not have permission to perform this action.',
}


class CollectionTestCase(ResolweAPITestCase):
    fixtures = ['users.yaml', 'collections.yaml', 'permissions.yaml']

    def setUp(self):
        self.collection1 = Collection.objects.get(pk=1)

        self.resource_name = 'collection'
        self.viewset = CollectionViewSet

        self.post_data = {
            'name': 'Test collection',
            'slug': 'test_collection',
        }

        super(CollectionTestCase, self).setUp()

    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(len(resp.data), 2)
        # self.assertEqual(len(resp.data), 3)

    def test_get_list_public_user(self):
        # public user
        resp = self._get_list()
        self.assertEqual(len(resp.data), 2)
        # check that you get right two objects
        self.assertEqual([p['id'] for p in resp.data], [1, 3])
        # check that (one of the) objects have expected keys
        self.assertKeys(resp.data[0], [u'slug', u'name', u'created', u'modified', u'contributor',
                                       u'description', u'id', u'settings', u'permissions',
                                       u'descriptor_schema', u'descriptor', u'data'])

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(len(resp.data), 3)

    def test_get_list_groups(self):
        resp = self._get_list(self.user3)
        self.assertEqual(len(resp.data), 1)
        # self.assertEqual(len(resp.data), 3)

    def test_post(self):
        # logged-in user
        collection_n = Collection.objects.count()
        resp = self._post(self.post_data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Collection.objects.count(), collection_n + 1)

        # public user
        collection_n += 1
        resp = self._post(self.post_data)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(Collection.objects.count(), collection_n)

    def test_get_detail(self):
        # public user w/ perm
        resp = self._get_detail(1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data[u'id'], 1)

        # public user w/o perm
        resp = self._get_detail(2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data[u'detail'], MESSAGES['NOT_FOUND'])

        # user w/ permissions
        resp = self._get_detail(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(resp.data, [u'slug', u'name', u'created', u'modified', u'contributor',
                                    u'description', u'settings', u'id', u'permissions',
                                    u'descriptor_schema', u'descriptor', u'data'])

        # user w/o permissions
        resp = self._get_detail(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data[u'detail'], MESSAGES['NOT_FOUND'])

    def test_patch(self):
        data = {'name': 'New collection'}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        p = Collection.objects.get(pk=1)
        self.assertEqual(p.name, 'New collection')

        # protected field
        data = {'created': '3042-01-01T09:00:00'}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        p = Collection.objects.get(pk=1)
        self.assertEqual(p.created.isoformat(), datetime(2015, 1, 1, 9, 0, 0).isoformat())

    def test_patch_no_perm(self):
        data = {'name': 'New collection'}
        resp = self._patch(2, data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        p = Collection.objects.get(pk=2)
        self.assertEqual(p.name, 'Test collection 2')

    def test_patch_public_user(self):
        data = {'name': 'New collection'}
        resp = self._patch(3, data)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        p = Collection.objects.get(pk=3)
        self.assertEqual(p.name, 'Test collection 3')

    def test_delete(self):
        resp = self._delete(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        q = Collection.objects.filter(pk=1).exists()
        self.assertFalse(q)

    def test_delete_no_perm(self):
        resp = self._delete(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        q = Collection.objects.filter(pk=2).exists()
        self.assertTrue(q)

    def test_delete_public_user(self):
        resp = self._delete(3)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        q = Collection.objects.filter(pk=3).exists()
        self.assertTrue(q)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from guardian.models import GroupObjectPermission, UserObjectPermission

from rest_framework import status

from .base import ResolweAPITestCase
from resolwe.flow.models import Collection
from resolwe.flow.views import CollectionViewSet


class PermissionsTestCase(ResolweAPITestCase):
    fixtures = ['users.yaml', 'collections.yaml', 'permissions.yaml']

    def setUp(self):
        self.collection1 = Collection.objects.get(pk=1)

        self.resource_name = 'collection'
        self.viewset = CollectionViewSet

        super(PermissionsTestCase, self).setUp()

    def test_add_permissions(self):
        data = {
            'users': {
                'add': {
                    2: ['download']
                }
            },
            'groups': {
                'add': {
                    1: ['view', 'edit']
                }
            }
        }

        # user w/o perms
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(2, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n)

        # add new permissions
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n + 1)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n + 2)

        # add already existing permissions
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n)

        # invalid permissions
        data = {'users': {'add': {2: ['delete']}}}
        user_perms_n = UserObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)

    def test_remove_permissions(self):
        data = {
            'users': {
                'remove': {
                    2: ['view', 'edit'],
                    42: ['view'],
                }
            },
            'groups': {
                'remove': {
                    1: ['share'],
                    42: ['edit'],
                }
            }
        }

        # user w/o perms
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(2, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n)

        # remove existing permissions
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n - 2)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n - 1)

        # remove non-existing permissions
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n)

        # invalid permissions
        data = {'users': {'remove': {2: ['delete']}}}
        user_perms_n = UserObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)

        # non-existing user
        data = {'users': {'remove': {42: ['edit']}}}
        user_perms_n = UserObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)

    def test_public_permissions(self):
        data = {
            'public': {
                'add': ['edit', 'share']
            },
        }
        user_perms_n = UserObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)

        data = {
            'public': {
                'remove': ['view']
            },
        }
        user_perms_n = UserObjectPermission.objects.count()
        resp = self._detail_permissions(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n - 1)

    def test_unauthorized_set(self):
        # registered user w/o perms
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        data = {'users': {'add': {2: ['download']}}}
        resp = self._detail_permissions(1, data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n)

        # public user
        user_perms_n = UserObjectPermission.objects.count()
        group_perms_n = GroupObjectPermission.objects.count()
        resp = self._detail_permissions(1, {'users': {'add': {2: ['view', 'edit']}}})
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_n)
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_n)



"""
Custom permissions for Flow API.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from rest_framework import permissions


class ResolwePermissions(permissions.DjangoObjectPermissions):
    """

    """
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.edit_%(model_name)s'],
        'PUT': ['%(app_label)s.edit_%(model_name)s'],
        'PATCH': ['%(app_label)s.edit_%(model_name)s'],
        'DELETE': ['%(app_label)s.edit_%(model_name)s'],
    }

    def has_permission(self, request, view):
        '''Always return `True` as we don't use model level perms.'''
        return True

    def has_object_permission(self, request, view, obj):
        # admins can do anything
        if request.user.is_superuser:
            return True

        # `share` permission is required for editing permissions
        if 'permissions' in (view.action or []):
            self.perms_map['POST'] = ['%(app_label)s.share_%(model_name)s']

        return super(ResolwePermissions, self).has_object_permission(request, view, obj)



from __future__ import unicode_literals

from itertools import chain

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from guardian.utils import get_identity
from guardian.utils import get_user_obj_perms_model
from guardian.utils import get_group_obj_perms_model
from guardian.compat import get_user_model


# Based on guardian.core.ObjectPermissionChecker.
def get_user_group_perms(user_or_group, obj):
    user, group = get_identity(user_or_group)

    if user and not user.is_active:
        return [], []
    User = get_user_model()
    ctype = ContentType.objects.get_for_model(obj)

    group_model = get_group_obj_perms_model(obj)
    group_rel_name = group_model.permission.field.related_query_name()
    if user:
        fieldname = '%s__group__%s' % (
            group_rel_name,
            User.groups.field.related_query_name(),
        )
        group_filters = {fieldname: user}
    else:
        group_filters = {'%s__group' % group_rel_name: group}
    if group_model.objects.is_generic():
        group_filters.update({
            '%s__content_type' % group_rel_name: ctype,
            '%s__object_pk' % group_rel_name: obj.pk,
        })
    else:
        group_filters['%s__content_object' % group_rel_name] = obj

    user_perms = []
    group_perms = []

    if user and user.is_superuser:
        user_perms = list(chain(*Permission.objects
                                .filter(content_type=ctype)
                                .values_list("codename")))
    elif user:
        model = get_user_obj_perms_model(obj)
        related_name = model.permission.field.related_query_name()
        user_filters = {'%s__user' % related_name: user}
        if model.objects.is_generic():
            user_filters.update({
                '%s__content_type' % related_name: ctype,
                '%s__object_pk' % related_name: obj.pk,
            })
        else:
            user_filters['%s__content_object' % related_name] = obj
        perms_qs = Permission.objects.filter(content_type=ctype)
        user_perms_qs = perms_qs.filter(**user_filters)
        user_perms = user_perms_qs.values_list("codename", flat=True)
        group_perms_qs = perms_qs.filter(**group_filters)
        group_perms = group_perms_qs.values_list("codename", flat=True)
    else:
        user_perms = list(set(chain(*Permission.objects
                                    .filter(content_type=ctype)
                                    .filter(**group_filters)
                                    .values_list("codename"))))

    return user_perms, group_perms



"""Utility functions"""


class BraceMessage(object):
    """Helper class that can be used to construct log messages with
    the new {}-string formatting syntax.

    NOTE: When using this helper class, one pays no signigicant
    performance penalty since the actual formatting only happens when
    (and if) the logged message is actually outputted to a log by a
    handler.

    Example usage:
    from genesis.utils.formatters import BraceMessage as __
    logger.error(__("Message with {0} {name}", 2, name="placeholders"))

    Source:
    https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles

    """

    def __init__(self, fmt, *args, **kwargs):
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.fmt.format(*self.args, **self.kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Open source dataflow package for Django framework.

See:
https://github.com/genialis/resolwe
"""

from setuptools import find_packages, setup
# Use codecs' open for a consistent encoding
from codecs import open
from os import path

base_dir = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(base_dir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Get package metadata from 'resolwe.__about__.py' file
about = {}
with open(path.join(base_dir, 'resolwe', '__about__.py'), encoding='utf-8') as f:
    exec(f.read(), about)

setup(
    name=about['__title__'],

    version=about['__version__'],

    description=about['__summary__'],
    long_description=long_description,

    url=about['__url__'],

    author=about['__author__'],
    author_email=about['__email__'],

    license=about['__license__'],

    packages=find_packages(exclude=['tests']),
    package_data={
        'resolwe': [
            'flow/static/flow/*.json',
            'flow/tests/processes/*.yml',
            'permissions/fixtures/*.yaml',
            'permissions/fixtures/readme.txt',
        ]
    },
    install_requires=[
        'django>=1.9,<1.10a1',
        # XXX: djangorestframework 3.3.3 fails to install on Read the Docs
        # Bug report: https://github.com/rtfd/readthedocs.org/issues/2101
        'djangorestframework==3.3.2',
        'djangorestframework-filters==0.8.0',
        'django-autoslug>=1.9.0',
        # XXX: django-guardian 1.4.0 is failing
        # 'django-guardian>=1.3.2',
        'django-guardian==1.3.2',
        'django-mathfilters>=0.3.0',
        'django-versionfield2>=0.4.0',
        'psycopg2>=2.5.0',
        'jsonfield>=1.0.3',
        'mock==1.3.0',
        'pyyaml>=3.11',
        'jsonschema>=2.4.0',
        'six>=1.10.0',
    ],
    extras_require = {
        'docs':  [
            'sphinx>=1.3.2',
            'sphinx_rtd_theme',
        ],
        'package': [
            'twine',
            'wheel',
        ],
        'test': [
            'django-jenkins>=0.17.0',
            'coverage>=3.7.1',
            'pep8>=1.6.2',
            'pylint>=1.4.3',
            'check-manifest',
            'readme_renderer',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',

        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',


        'License :: OSI Approved :: Apache Software License',

        'Operating System :: OS Independent',

        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='resolwe dataflow django',
)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase

from resolwe.flow.managers import manager
from resolwe.flow.models import Data, Process


class BackendTest(TestCase):
    def setUp(self):
        u = get_user_model().objects.create_superuser('test', 'test@genialis.com', 'test')
        self.p = Process(slug='test-processor',
                         name='Test Process',
                         contributor=u,
                         type='data:test',
                         version=1)
        self.p.save()

        self.d = Data(slug='test-data',
                      name='Test Data',
                      contributor=u,
                      process=self.p)
        self.d.save()

    def tearDown(self):
        for data in Data.objects.all():
            data_dir = os.path.join(settings.FLOW_EXECUTOR['DATA_PATH'], str(data.id))
            shutil.rmtree(data_dir, ignore_errors=True)

    def test_manager(self):
        manager.communicate(verbosity=0)

    def test_dtlbash(self):
        self.p.slug = 'test-processor-dtlbash'
        self.p.run = {'script': """
gen-info \"Test processor info\"
gen-warning \"Test processor warning\"

echo '{"proc.info": "foo"}'
"""}
        self.p.save()

        self.d.slug = 'test-data-dtlbash'
        self.d.process = self.p
        self.d.save()
        self.d = Data(id=self.d.id)
